@@ -1,34 +1,40 @@
 """Provides the main() routine and ColaApplication"""
 from __future__ import division, absolute_import, unicode_literals
+from functools import partial
 import argparse
 import os
 import signal
 import sys
+import time
 
 __copyright__ = """
-Copyright (C) 2009-2017 David Aguilar and contributors
+Copyright (C) 2007-2017 David Aguilar and contributors
 """
 
-from . import core
 try:
     from qtpy import QtCore
 except ImportError:
-    errmsg = """
-Sorry, you do not seem to have PyQt5, Pyside, or PyQt4 installed.
-Please install it before using git-cola, e.g.:
-    $ sudo apt-get install python-qt4
-"""
-    core.error(errmsg)
+    sys.stderr.write("""
+You do not seem to have PyQt5, PySide, or PyQt4 installed.
+Please install it before using git-cola, e.g. on a Debian/Ubutnu system:
+
+    sudo apt-get install python-pyqt5 python-pyqt5.qtwebkit
+
+""")
+    sys.exit(1)
 
 from qtpy import QtGui
 from qtpy import QtWidgets
+from qtpy.QtCore import Qt
 
 # Import cola modules
 from .i18n import N_
 from .interaction import Interaction
 from .models import main
+from .models import selection
 from .widgets import cfgactions
 from .widgets import defs
+from .widgets import standard
 from .widgets import startup
 from .settings import Session
 from . import cmds
@@ -37,15 +43,18 @@ from . import compat
 from . import fsmonitor
 from . import git
 from . import gitcfg
+from . import guicmds
 from . import icons
 from . import i18n
 from . import qtcompat
 from . import qtutils
 from . import resources
+from . import utils
 from . import version
 
 
 def setup_environment():
+    """Set environment variables to control git's behavior"""
     # Allow Ctrl-C to exit
     signal.signal(signal.SIGINT, signal.SIG_DFL)
 
@@ -119,7 +128,7 @@ def setup_environment():
     compat.setenv('GIT_MERGE_AUTOEDIT', 'no')
 
 
-def get_icon_themes():
+def get_icon_themes(context):
     """Return the default icon theme names"""
     themes = []
 
@@ -127,7 +136,7 @@ def get_icon_themes():
     if icon_themes_env:
         themes.extend([x for x in icon_themes_env.split(':') if x])
 
-    icon_themes_cfg = gitcfg.current().get_all('cola.icontheme')
+    icon_themes_cfg = context.cfg.get_all('cola.icontheme')
     if icon_themes_cfg:
         themes.extend(icon_themes_cfg)
 
@@ -144,21 +153,21 @@ class ColaApplication(object):
     ColaApplication handles i18n of user-visible data
     """
 
-    def __init__(self, argv, locale=None, gui=True, icon_themes=None):
+    def __init__(self, context, argv, locale=None, icon_themes=None):
         cfgactions.install()
         i18n.install(locale)
         qtcompat.install()
-        qtutils.install()
-        icons.install(icon_themes or get_icon_themes())
+        guicmds.install()
+        standard.install()
+        icons.install(icon_themes or get_icon_themes(context))
 
-        if gui:
-            self._app = ColaQApplication(list(argv))
-            self._app.setWindowIcon(icons.cola())
-            self._install_style()
-        else:
-            self._app = QtCore.QCoreApplication(argv)
+        self.context = context
+        self._app = ColaQApplication(context, list(argv))
+        self._app.setWindowIcon(icons.cola())
+        self._install_style()
 
     def _install_style(self):
+        """Generate and apply a stylesheet to the app"""
         palette = self._app.palette()
         window = palette.color(QtGui.QPalette.Window)
         highlight = palette.color(QtGui.QPalette.Highlight)
@@ -228,72 +237,77 @@ class ColaApplication(object):
                        radio_size=defs.checkbox))
 
     def activeWindow(self):
-        """Wrap activeWindow()"""
+        """QApplication::activeWindow() pass-through"""
         return self._app.activeWindow()
 
     def desktop(self):
+        """QApplication::desktop() pass-through"""
         return self._app.desktop()
 
     def start(self):
         """Wrap exec_() and start the application"""
         # Defer connection so that local cola.inotify is honored
-        fsmonitor.current().files_changed.connect(self._update_files)
+        context = self.context
+        monitor = context.fsmonitor
+        monitor.files_changed.connect(
+            cmds.run(cmds.Refresh, context), type=Qt.QueuedConnection)
+        monitor.config_changed.connect(
+            cmds.run(cmds.RefreshConfig, context), type=Qt.QueuedConnection)
         # Start the filesystem monitor thread
-        fsmonitor.current().start()
+        monitor.start()
         return self._app.exec_()
 
     def stop(self):
         """Finalize the application"""
-        fsmonitor.current().stop()
+        self.context.fsmonitor.stop()
         # Workaround QTBUG-52988 by deleting the app manually to prevent a
         # crash during app shutdown.
         # https://bugreports.qt.io/browse/QTBUG-52988
         try:
             del self._app
-        except:
+        except (AttributeError, RuntimeError):
             pass
         self._app = None
 
-    def set_view(self, view):
-        if hasattr(self._app, 'view'):
-            self._app.view = view
-
-    def set_context(self, context):
-        if hasattr(self._app, 'context'):
-            self._app.context = context
-
-    def _update_files(self):
-        # Respond to file system updates
-        cmds.do(cmds.Refresh)
+    def exit(self, status):
+        """QApplication::exit(status) pass-through"""
+        return self._app.exit(status)
 
 
 class ColaQApplication(QtWidgets.QApplication):
+    """QApplication implementation for handling custom events"""
 
-    def __init__(self, argv):
+    def __init__(self, context, argv):
         super(ColaQApplication, self).__init__(argv)
-        self.view = None  # injected by application_start()
-        self.context = None  # injected by application_start()
+        self.context = context
 
     def event(self, e):
+        """Respond to focus events for the cola.refreshonfocus feature"""
         if e.type() == QtCore.QEvent.ApplicationActivate:
-            cfg = gitcfg.current()
-            if (self.context and self.context.model.git.is_valid()
-                    and cfg.get('cola.refreshonfocus', False)):
-                cmds.do(cmds.Refresh)
+            context = self.context
+            if context:
+                cfg = context.cfg
+                if (context.git.is_valid()
+                        and cfg.get('cola.refreshonfocus', default=False)):
+                    cmds.do(cmds.Refresh, context)
         return super(ColaQApplication, self).event(e)
 
     def commitData(self, session_mgr):
         """Save session data"""
-        if self.view is None:
+        if not self.context or not self.context.view:
+            return
+        view = self.context.view
+        if not hasattr(view, 'save_state'):
             return
         sid = session_mgr.sessionId()
         skey = session_mgr.sessionKey()
         session_id = '%s_%s' % (sid, skey)
         session = Session(session_id, repo=core.getcwd())
-        self.view.save_state(settings=session)
+        view.save_state(settings=session)
 
 
 def process_args(args):
+    """Process and verify command-line arguments"""
     if args.version:
         # Accept 'git cola --version' or 'git cola version'
         version.print_version()
@@ -310,11 +324,12 @@ def process_args(args):
     if not core.isdir(repo):
         errmsg = N_('fatal: "%s" is not a directory.  '
                     'Please specify a correct --repo <path>.') % repo
-        core.stderr(errmsg)
+        core.print_stderr(errmsg)
         sys.exit(core.EXIT_USAGE)
 
 
 def restore_session(args):
+    """Load a session based on the window-manager provided arguments"""
     # args.settings is provided when restoring from a session.
     args.settings = None
     if args.session is None:
@@ -333,28 +348,41 @@ def application_init(args, update=False):
     setup_environment()
     process_args(args)
 
-    app = new_application(args)
-    model = new_model(app, args.repo,
-                      prompt=args.prompt, settings=args.settings)
+    context = new_context(args)
+    timer = context.timer
+    timer.start('init')
+
+    new_worktree(context, args.repo, args.prompt, args.settings)
+
     if update:
-        model.update_status()
-    cfg = gitcfg.current()
-    return ApplicationContext(args, app, cfg, model)
+        context.model.update_status()
+
+    timer.stop('init')
+    if args.perf:
+        timer.display('init')
+    return context
 
 
-def context_init(context, view):
-    """Store the view for session management"""
-    context.app.set_view(view)
-    context.app.set_context(context)
-    # Make sure that we start out on top
-    view.show()
-    view.raise_()
+def new_context(args):
+    """Create top-level ApplicationContext objects"""
+    context = ApplicationContext(args)
+    context.git = git.create()
+    context.cfg = gitcfg.create(context)
+    context.fsmonitor = fsmonitor.create(context)
+    context.selection = selection.create()
+    context.model = main.create(context)
+    context.app = new_application(context, args)
+    context.timer = Timer()
+
+    return context
 
 
 def application_run(context, view, start=None, stop=None):
     """Run the application main loop"""
-    # Initialize and run startup callbacks
-    context_init(context, view)
+    context.set_view(view)
+    view.show()
+
+    # Startup callbacks
     if start:
         start(context, view)
     # Start the event loop
@@ -370,22 +398,23 @@ def application_run(context, view, start=None, stop=None):
 def application_start(context, view):
     """Show the GUI and start the main event loop"""
     # Store the view for session management
-    return application_run(context, view, start=start, stop=stop)
+    return application_run(context, view,
+                           start=default_start, stop=default_stop)
 
 
-def start(context, view):
+def default_start(context, _view):
     """Scan for the first time"""
-    runtask = qtutils.RunTask(parent=view)
-    init_update_task(view, runtask, context.model)
-    QtCore.QTimer.singleShot(0, _send_msg)
+    QtCore.QTimer.singleShot(0, startup_message)
+    QtCore.QTimer.singleShot(0, lambda: async_update(context))
 
 
-def stop(context, view):
+def default_stop(_context, _view):
     """All done, cleanup"""
     QtCore.QThreadPool.globalInstance().waitForDone()
 
 
 def add_common_arguments(parser):
+    """Add command arguments to the ArgumentParser"""
     # We also accept 'git cola version'
     parser.add_argument('--version', default=False, action='store_true',
                         help='print version number')
@@ -407,22 +436,30 @@ def add_common_arguments(parser):
     parser.add_argument('-session', metavar='<session>', default=None,
                         help=argparse.SUPPRESS)
 
-
-def new_application(args):
-    # Initialize the app
-    return ColaApplication(sys.argv, icon_themes=args.icon_themes)
+    # Enable timing information
+    parser.add_argument('--perf', action='store_true', default=False,
+                        help=argparse.SUPPRESS)
 
 
-def new_model(app, repo, prompt=False, settings=None):
-    model = main.model()
+def new_application(context, args):
+    """Create a new ColaApplication"""
+    return ColaApplication(context, sys.argv, icon_themes=args.icon_themes)
+
+
+def new_worktree(context, repo, prompt, settings):
+    """Find a Git repository, or prompt for one when not found"""
+    model = context.model
+    cfg = context.cfg
+    parent = qtutils.active_window()
     valid = False
+
     if not prompt:
         valid = model.set_worktree(repo)
         if not valid:
             # We are not currently in a git repository so we need to find one.
             # Before prompting the user for a repository, check if they've
             # configured a default repository and attempt to use it.
-            default_repo = gitcfg.current().get('cola.defaultrepo')
+            default_repo = cfg.get('cola.defaultrepo')
             if default_repo:
                 valid = model.set_worktree(default_repo)
 
@@ -430,31 +467,26 @@ def new_model(app, repo, prompt=False, settings=None):
         # If we've gotten into this loop then that means that neither the
         # current directory nor the default repository were available.
         # Prompt the user for a repository.
-        startup_dlg = startup.StartupDialog(app.activeWindow(),
-                                            settings=settings)
+        startup_dlg = startup.StartupDialog(context, parent, settings=settings)
         gitdir = startup_dlg.find_git_repo()
         if not gitdir:
             sys.exit(core.EXIT_NOINPUT)
         valid = model.set_worktree(gitdir)
 
-    return model
 
-
-def init_update_task(parent, runtask, model):
+def async_update(context):
     """Update the model in the background
 
     git-cola should startup as quickly as possible.
 
     """
-
-    def update_status():
-        model.update_status(update_index=True)
-
-    task = qtutils.SimpleTask(parent, update_status)
-    runtask.start(task)
+    update_status = partial(context.model.update_status, update_index=True)
+    task = qtutils.SimpleTask(context.view, update_status)
+    context.runtask.start(task)
 
 
-def _send_msg():
+def startup_message():
+    """Print debug startup messages"""
     trace = git.GIT_COLA_TRACE
     if trace == '2' or trace == 'trace':
         msg1 = 'info: debug level 2: trace mode enabled'
@@ -468,10 +500,93 @@ def _send_msg():
         Interaction.log(msg2)
 
 
-class ApplicationContext(object):
+class Timer(object):
+    """Simple performance timer"""
 
-    def __init__(self, args, app, cfg, model):
+    def __init__(self):
+        self._data = {}
+
+    def start(self, key):
+        """Start a timer"""
+        now = time.time()
+        self._data[key] = [now, now]
+
+    def stop(self, key):
+        """Stop a timer and return its elapsed time"""
+        entry = self._data[key]
+        entry[1] = time.time()
+        return self.elapsed(key)
+
+    def elapsed(self, key):
+        """Return the elapsed time for a timer"""
+        entry = self._data[key]
+        return entry[1] - entry[0]
+
+    def display(self, key):
+        """Display a timer"""
+        elapsed = self.elapsed(key)
+        sys.stdout.write('%s: %.5fs\n' % (key, elapsed))
+
+
+class ApplicationContext(object):
+    """Context for performing operations on Git and related data models"""
+
+    def __init__(self, args):
         self.args = args
-        self.app = app
-        self.cfg = cfg
-        self.model = model
+        self.app = None  # ColaApplication
+        self.git = None  # git.Git
+        self.cfg = None  # gitcfg.GitConfig
+        self.model = None  # main.MainModel
+        self.timer = None  # Timer
+        self.runtask = None  # qtutils.RunTask
+        self.selection = None  # selection.SelectionModel
+        self.fsmonitor = None  # fsmonitor
+        self.view = None  # QWidget
+
+    def set_view(self, view):
+        """Initialize view-specific members"""
+        self.view = view
+        self.runtask = qtutils.RunTask(parent=view)
+
+
+def winmain(main_fn, *argv):
+    """Find Git and launch main(argv)"""
+    git_path = find_git()
+    if git_path:
+        prepend_path(git_path)
+    return main_fn(*argv)
+
+
+def find_git():
+    """Return the path of git.exe, or None if we can't find it."""
+    if not utils.is_win32():
+        return None  # UNIX systems have git in their $PATH
+
+    # If the user wants to use a Git/bin/ directory from a non-standard
+    # directory then they can write its location into
+    # ~/.config/git-cola/git-bindir
+    git_bindir = os.path.expanduser(os.path.join('~', '.config', 'git-cola',
+                                                 'git-bindir'))
+    if core.exists(git_bindir):
+        custom_path = core.read(git_bindir).strip()
+        if custom_path and core.exists(custom_path):
+            return custom_path
+
+    # Try to find Git's bin/ directory in one of the typical locations
+    pf = os.environ.get('ProgramFiles', 'C:\\Program Files')
+    pf32 = os.environ.get('ProgramFiles(x86)', 'C:\\Program Files (x86)')
+    for p in [pf32, pf, 'C:\\']:
+        candidate = os.path.join(p, 'Git\\bin')
+        if os.path.isdir(candidate):
+            return candidate
+
+    return None
+
+
+def prepend_path(path):
+    """Adds git to the PATH.  This is needed on Windows."""
+    path = core.decode(path)
+    path_entries = core.getenv('PATH', '').split(os.pathsep)
+    if path not in path_entries:
+        path_entries.insert(0, path)
+        compat.setenv('PATH', os.pathsep.join(path_entries))
