@@ -19,8 +19,8 @@ from .compat import PY3
 from .compat import WIN32
 
 # /usr/include/stdlib.h
-#define EXIT_SUCCESS    0   /* Successful exit status.  */
-#define EXIT_FAILURE    1   /* Failing exit status.  */
+# #define EXIT_SUCCESS    0   /* Successful exit status.  */
+# #define EXIT_FAILURE    1   /* Failing exit status.  */
 EXIT_SUCCESS = 0
 EXIT_FAILURE = 1
 
@@ -32,10 +32,13 @@ EXIT_USAGE = 64
 EXIT_NOINPUT = 66
 EXIT_UNAVAILABLE = 69
 
+# Default encoding
+ENCODING = 'utf-8'
+
 # Some files are not in UTF-8; some other aren't in any codification.
 # Remember that GIT doesn't care about encodings (saves binary data)
 _encoding_tests = [
-    'utf-8',
+    ENCODING,
     'iso-8859-15',
     'windows1252',
     'ascii',
@@ -73,7 +76,7 @@ def decode(value, encoding=None, errors='strict'):
     if value is None:
         result = None
     elif isinstance(value, ustr):
-        result = UStr(value, 'utf-8')
+        result = UStr(value, ENCODING)
     else:
         result = None
         if encoding is None:
@@ -81,17 +84,17 @@ def decode(value, encoding=None, errors='strict'):
         else:
             encoding_tests = itertools.chain([encoding], _encoding_tests)
 
-        for encoding in encoding_tests:
+        for enc in encoding_tests:
             try:
-                decoded = value.decode(encoding, errors)
-                result = UStr(decoded, encoding)
+                decoded = value.decode(enc, errors)
+                result = UStr(decoded, enc)
                 break
             except ValueError:
                 pass
 
         if result is None:
-            encoding = 'utf-8'
-            result = value.decode(encoding, errors='ignore')
+            decoded = value.decode(ENCODING, errors='ignore')
+            result = UStr(decoded, ENCODING)
 
     return result
 
@@ -99,9 +102,9 @@ def decode(value, encoding=None, errors='strict'):
 def encode(string, encoding=None):
     """encode(unencoded_string) returns a string encoded in utf-8
     """
-    if type(string) is not ustr:
+    if not isinstance(string, ustr):
         return string
-    return string.encode(encoding or 'utf-8', 'replace')
+    return string.encode(encoding or ENCODING, 'replace')
 
 
 def mkpath(path, encoding=None):
@@ -230,7 +233,7 @@ def communicate(proc):
     return proc.communicate()
 
 
-def run_command(cmd, encoding=None, *args, **kwargs):
+def run_command(cmd, *args, **kwargs):
     """Run the given command to completion, and return its results.
 
     This provides a simpler interface to the subprocess module.
@@ -238,12 +241,15 @@ def run_command(cmd, encoding=None, *args, **kwargs):
     The other arguments are passed on to start_command().
 
     """
+    encoding = kwargs.pop('encoding', None)
     process = start_command(cmd, *args, **kwargs)
     (output, errors) = communicate(process)
     output = decode(output, encoding=encoding)
     errors = decode(errors, encoding=encoding)
     exit_code = process.returncode
-    return (exit_code, output or '', errors or '')
+    return (exit_code,
+            output or UStr('', ENCODING),
+            errors or UStr('', ENCODING))
 
 
 @interruptable
@@ -346,22 +352,22 @@ def xopen(path, mode='r', encoding=None):
     return open(mkpath(path, encoding=encoding), mode)
 
 
-def stdout(msg, linesep='\n'):
+def print_stdout(msg, linesep='\n'):
     msg = msg + linesep
     if PY2:
-        msg = encode(msg, encoding='utf-8')
+        msg = encode(msg, encoding=ENCODING)
     sys.stdout.write(msg)
 
 
-def stderr(msg, linesep='\n'):
+def print_stderr(msg, linesep='\n'):
     msg = msg + linesep
     if PY2:
-        msg = encode(msg, encoding='utf-8')
+        msg = encode(msg, encoding=ENCODING)
     sys.stderr.write(msg)
 
 
 def error(msg, status=EXIT_FAILURE, linesep='\n'):
-    stderr(msg, linesep=linesep)
+    print_stderr(msg, linesep=linesep)
     sys.exit(status)
 
 
@@ -375,9 +381,46 @@ chdir = wrap(mkpath, os.chdir)
 exists = wrap(mkpath, os.path.exists)
 expanduser = wrap(encode, os.path.expanduser, decorator=decode)
 if PY2:
-    getcwd = decorate(decode, os.getcwd)
+    if hasattr(os, 'getcwdu'):
+        getcwd = os.getcwdu
+    else:
+        getcwd = decorate(decode, os.getcwd)
 else:
     getcwd = os.getcwd
+
+
+# NOTE: find_executable() is originally from the stdlib, but starting with
+# python3.7 the stdlib no longer bundles distutils.
+def _find_executable(executable, path=None):
+    """Tries to find 'executable' in the directories listed in 'path'.
+
+    A string listing directories separated by 'os.pathsep'; defaults to
+    os.environ['PATH'].  Returns the complete filename or None if not found.
+    """
+    if path is None:
+        path = os.environ['PATH']
+
+    paths = path.split(os.pathsep)
+    _, ext = os.path.splitext(executable)
+
+    if (sys.platform == 'win32') and (ext != '.exe'):
+        executable = executable + '.exe'
+
+    if not os.path.isfile(executable):
+        for p in paths:
+            f = os.path.join(p, executable)
+            if os.path.isfile(f):
+                # the file exists, we have a shot at spawn working
+                return f
+        return None
+
+    return executable
+
+
+if PY2:
+    find_executable = wrap(mkpath, _find_executable, decorator=decode)
+else:
+    find_executable = wrap(decode, _find_executable, decorator=decode)
 isdir = wrap(mkpath, os.path.isdir)
 isfile = wrap(mkpath, os.path.isfile)
 islink = wrap(mkpath, os.path.islink)
