@@ -16,23 +16,20 @@ from .widgets import filetree
 from .widgets import standard
 
 
-def diff_commits(parent, a, b):
+def diff_commits(context, parent, a, b):
     """Show a dialog for diffing two commits"""
-    dlg = Difftool(parent, a=a, b=b)
+    dlg = Difftool(context, parent, a=a, b=b)
     dlg.show()
     dlg.raise_()
     return dlg.exec_() == QtWidgets.QDialog.Accepted
 
 
-def diff_expression(parent, expr,
-                    create_widget=False,
-                    hide_expr=False,
-                    focus_tree=False):
+def diff_expression(context, parent, expr,
+                    create_widget=False, hide_expr=False, focus_tree=False):
     """Show a diff dialog for diff expressions"""
-    dlg = Difftool(parent,
-                         expr=expr,
-                         hide_expr=hide_expr,
-                         focus_tree=focus_tree)
+    dlg = Difftool(context, parent,
+                   expr=expr, hide_expr=hide_expr,
+                   focus_tree=focus_tree)
     if create_widget:
         return dlg
     dlg.show()
@@ -42,12 +39,13 @@ def diff_expression(parent, expr,
 
 class Difftool(standard.Dialog):
 
-    def __init__(self, parent, a=None, b=None, expr=None, title=None,
+    def __init__(self, context, parent, a=None, b=None, expr=None, title=None,
                  hide_expr=False, focus_tree=False):
         """Show files with differences and launch difftool"""
 
         standard.Dialog.__init__(self, parent=parent)
 
+        self.context = context
         self.a = a
         self.b = b
         self.diff_expr = expr
@@ -58,7 +56,7 @@ class Difftool(standard.Dialog):
         self.setWindowTitle(title)
         self.setWindowModality(Qt.WindowModal)
 
-        self.expr = completion.GitRefLineEdit(parent=self)
+        self.expr = completion.GitRefLineEdit(context, parent=self)
         if expr is not None:
             self.expr.setText(expr)
 
@@ -147,10 +145,11 @@ class Difftool(standard.Dialog):
         self.refresh_filenames()
 
     def refresh_filenames(self):
+        context = self.context
         if self.a and self.b is None:
-            filenames = gitcmds.diff_index_filenames(self.a)
+            filenames = gitcmds.diff_index_filenames(context, self.a)
         else:
-            filenames = gitcmds.diff(self.diff_arg)
+            filenames = gitcmds.diff(context, self.diff_arg)
         self.tree.set_filenames(filenames, select=True)
 
     def tree_selection_changed(self):
@@ -158,15 +157,16 @@ class Difftool(standard.Dialog):
         self.diff_button.setEnabled(has_selection)
         self.diff_all_button.setEnabled(has_selection)
 
-    def tree_double_clicked(self, item, column):
+    def tree_double_clicked(self, item, _column):
         path = self.tree.filename_from_item(item)
         left, right = self._left_right_args()
-        cmds.difftool_launch(left=left, right=right, paths=[path])
+        cmds.difftool_launch(
+            self.context, left=left, right=right, paths=[path])
 
     def diff(self, dir_diff=False):
         paths = self.tree.selected_filenames()
         left, right = self._left_right_args()
-        cmds.difftool_launch(left=left, right=right, paths=paths,
+        cmds.difftool_launch(self.context, left=left, right=right, paths=paths,
                              dir_diff=dir_diff)
 
     def _left_right_args(self):
@@ -182,4 +182,4 @@ class Difftool(standard.Dialog):
 
     def edit(self):
         paths = self.tree.selected_filenames()
-        cmds.do(cmds.Edit, paths)
+        cmds.do(cmds.Edit, self.context, paths)
