@@ -1,27 +1,37 @@
 """i18n and l10n support for git-cola"""
-
+from __future__ import division, absolute_import, unicode_literals
 import gettext as _gettext
 import os
 import sys
 
-from cola import compat
-from cola import core
-from cola import resources
-from cola import xdg
+from . import compat
+from . import core
+from . import resources
 
 _null_translation = _gettext.NullTranslations()
 _translation = _null_translation
 
 
 def gettext(s):
-    txt = _translation.ugettext(s)
-    if txt[-6:-4] == '@@': # handle @@verb / @@noun
+    try:
+        txt = _translation.ugettext(s)
+    except AttributeError:
+        # Python 3 compat
+        _translation.ugettext = _translation.gettext
+        txt = _translation.gettext(s)
+    if txt[-6:-4] == '@@':  # handle @@verb / @@noun
         txt = txt[:-6]
     return txt
 
 
 def ngettext(s, p, n):
-    return _translation.ungettext(s, p, n)
+    try:
+        txt = _translation.ungettext(s, p, n)
+    except AttributeError:
+        # Python 3 compat
+        _translation.ungettext = _translation.ngettext
+        txt = _translation.ngettext(s, p, n)
+    return txt
 
 
 def N_(s):
@@ -29,10 +39,12 @@ def N_(s):
 
 
 def install(locale):
+    # pylint: disable=global-statement
     global _translation
     if sys.platform == 'win32':
         _check_win32_locale()
     if locale:
+        compat.setenv('LANGUAGE', locale)
         compat.setenv('LANG', locale)
         compat.setenv('LC_MESSAGES', locale)
     _install_custom_language()
@@ -41,7 +53,9 @@ def install(locale):
                                         localedir=_get_locale_dir(),
                                         fallback=True)
 
+
 def uninstall():
+    # pylint: disable=global-statement
     global _translation
     _translation = _null_translation
 
@@ -52,19 +66,19 @@ def _get_locale_dir():
 
 def _install_custom_language():
     """Allow a custom language to be set in ~/.config/git-cola/language"""
-    lang_file = xdg.config_home('language')
+    lang_file = resources.config_home('language')
     if not core.exists(lang_file):
         return
     try:
         lang = core.read(lang_file).strip()
-    except:
+    except (OSError, IOError):
         return
     if lang:
         compat.setenv('LANGUAGE', lang)
 
 
 def _check_win32_locale():
-    for i in ('LANGUAGE','LC_ALL','LC_MESSAGES','LANG'):
+    for i in ('LANGUAGE', 'LC_ALL', 'LC_MESSAGES', 'LANG'):
         if os.environ.get(i):
             break
     else:
