@@ -1,36 +1,36 @@
 """Provides a selection model to handle selection."""
+from __future__ import division, absolute_import, unicode_literals
 
 import collections
 
-from cola.compat import set
-from cola.observable import Observable
-from cola.decorators import memoize
+from ..observable import Observable
 
 State = collections.namedtuple('State', 'staged unmerged modified untracked')
 
 
-@memoize
-def selection_model():
-    """Provides access to a static SelectionModel instance."""
+def create():
+    """Create a SelectionModel"""
     return SelectionModel()
 
 
-def selection():
-    """Return the current selection."""
-    return selection_model().selection()
+def pick(s):
+    """Choose the first list from stage, unmerged, modified, untracked"""
+    if s.staged:
+        files = s.staged
+    elif s.unmerged:
+        files = s.unmerged
+    elif s.modified:
+        files = s.modified
+    elif s.untracked:
+        files = s.untracked
+    else:
+        files = []
+    return files
 
 
-def single_selection():
-    """Scan across staged, modified, etc. and return a single item."""
-    return selection_model().single_selection()
-
-
-def selected_group():
-    return selection_model().group()
-
-
-def filename():
-    return selection_model().filename()
+def union(s):
+    """Return the union of all selected items in a sorted list"""
+    return list(sorted(set(s.staged + s.unmerged + s.modified + s.untracked)))
 
 
 def _filter(a, b):
@@ -49,9 +49,8 @@ class SelectionModel(Observable):
 
     # These properties wrap the individual selection items
     # to provide higher-level pseudo-selections.
-    unstaged = property(lambda self: self.unmerged +
-                                     self.modified +
-                                     self.untracked)
+    unstaged = property(lambda self:
+                        self.unmerged + self.modified + self.untracked)
 
     def __init__(self):
         Observable.__init__(self)
@@ -90,6 +89,7 @@ class SelectionModel(Observable):
                      self.modified, self.untracked)
 
     def single_selection(self):
+        """Scan across staged, modified, etc. and return a single item."""
         st = None
         m = None
         um = None
@@ -107,20 +107,18 @@ class SelectionModel(Observable):
     def filename(self):
         paths = [p for p in self.single_selection() if p is not None]
         if paths:
-            return paths[0]
+            filename = paths[0]
         else:
-            return None
+            filename = None
+        return filename
 
     def group(self):
         """A list of selected files in various states of being"""
-        selection = []
-        s = self.selection()
-        if s.staged:
-            selection = s.staged
-        elif s.unmerged:
-            selection = s.unmerged
-        elif s.modified:
-            selection = s.modified
-        elif s.untracked:
-            selection = s.untracked
-        return selection
+        return pick(self.selection())
+
+    def pick(self, selection):
+        return pick(selection)
+
+    def union(self):
+        """Return the union of all selected items in a sorted list"""
+        return union(self)
