@@ -11,24 +11,30 @@ from . import resources
 
 
 def mkdict(obj):
-    if type(obj) is dict:
-        return obj
+    """Transform None and non-dicts into dicts"""
+    if isinstance(obj, dict):
+        value = obj
     else:
-        return {}
+        value = {}
+    return value
 
 
 def mklist(obj):
-    if type(obj) is list:
-        return obj
+    """Transform None and non-lists into lists"""
+    if isinstance(obj, list):
+        value = obj
+    elif isinstance(obj, tuple):
+        value = list(obj)
     else:
-        return []
+        value = []
+    return value
 
 
 def read_json(path):
     try:
         with core.xopen(path, 'rt') as fp:
             return mkdict(json.load(fp))
-    except:  # bad path or json
+    except (ValueError, TypeError, OSError, IOError):  # bad path or json
         return {}
 
 
@@ -39,22 +45,24 @@ def write_json(values, path):
             core.makedirs(parent)
         with core.xopen(path, 'wt') as fp:
             json.dump(values, fp, indent=4)
-    except:
+    except (ValueError, TypeError, OSError, IOError):
         sys.stderr.write('git-cola: error writing "%s"\n' % path)
 
 
 class Settings(object):
-    _file = resources.config_home('settings')
+    config_path = resources.config_home('settings')
     bookmarks = property(lambda self: mklist(self.values['bookmarks']))
     gui_state = property(lambda self: mkdict(self.values['gui_state']))
     recent = property(lambda self: mklist(self.values['recent']))
+    copy_formats = property(lambda self: mklist(self.values['copy_formats']))
 
     def __init__(self, verify=git.is_git_worktree):
         """Load existing settings if they exist"""
         self.values = {
-                'bookmarks': [],
-                'gui_state': {},
-                'recent': [],
+            'bookmarks': [],
+            'gui_state': {},
+            'recent': [],
+            'copy_formats': [],
         }
         self.verify = verify
 
@@ -69,7 +77,7 @@ class Settings(object):
         for bookmark in missing_bookmarks:
             try:
                 self.bookmarks.remove(bookmark)
-            except:
+            except ValueError:
                 pass
 
         for recent in self.recent:
@@ -79,7 +87,7 @@ class Settings(object):
         for recent in missing_recent:
             try:
                 self.recent.remove(recent)
-            except:
+            except ValueError:
                 pass
 
     def add_bookmark(self, path, name):
@@ -112,7 +120,7 @@ class Settings(object):
     def rename_bookmark(self, path, name, new_name):
         return self.rename_entry(self.bookmarks, path, name, new_name)
 
-    def add_recent(self, path):
+    def add_recent(self, path, max_recent):
         try:
             index = [recent['path'] for recent in self.recent].index(path)
             entry = self.recent.pop(index)
@@ -122,14 +130,15 @@ class Settings(object):
                 'path': path,
             }
         self.recent.insert(0, entry)
-
-        if len(self.recent) >= 8:
+        if len(self.recent) >= max_recent:
             self.recent.pop()
 
     def remove_recent(self, path):
         """Removes an item from the recent items list"""
         try:
-            index = [recent.get('path', '') for recent in self.recent].index(path)
+            index = [
+                recent.get('path', '') for recent in self.recent
+            ].index(path)
         except ValueError:
             return
         try:
@@ -141,7 +150,7 @@ class Settings(object):
         return self.rename_entry(self.recent, path, name, new_name)
 
     def path(self):
-        return self._file
+        return self.config_path
 
     def save(self):
         write_json(self.values, self.path())
@@ -156,12 +165,12 @@ class Settings(object):
         # Upgrade bookmarks to the new dict-based bookmarks format.
         if self.bookmarks and not isinstance(self.bookmarks[0], dict):
             bookmarks = [dict(name=os.path.basename(path), path=path)
-                            for path in self.bookmarks]
+                         for path in self.bookmarks]
             self.values['bookmarks'] = bookmarks
 
         if self.recent and not isinstance(self.recent[0], dict):
             recent = [dict(name=os.path.basename(path), path=path)
-                        for path in self.recent]
+                      for path in self.recent]
             self.values['recent'] = recent
 
     def asdict(self):
@@ -221,7 +230,7 @@ class Session(Settings):
             self.values.update(read_json(path))
             try:
                 os.unlink(path)
-            except:
+            except (OSError, ValueError):
                 pass
             return True
         return False
