@@ -1,7 +1,7 @@
-# Copyright (c) 2008 David Aguilar
+# Copyright (C) 2007-2018 David Aguilar and contributors
 """This module provides miscellaneous utility functions."""
 from __future__ import division, absolute_import, unicode_literals
-
+import copy
 import os
 import random
 import re
@@ -11,12 +11,29 @@ import tempfile
 import time
 import traceback
 
-from cola import core
-from cola.decorators import memoize
+from . import core
 
 random.seed(hash(time.time()))
 
 
+def asint(obj, default=0):
+    """Make any value into an int, even if the cast fails"""
+    try:
+        value = int(obj)
+    except TypeError:
+        value = default
+    return value
+
+
+def clamp(value, lo, hi):
+    """Clamp a value to the specified range"""
+    return min(hi, max(lo, value))
+
+
+def epoch_millis():
+    return int(time.time() * 1000)
+
+
 def add_parents(paths):
     """Iterate over each item in the set and add its parent directories."""
     all_paths = set()
@@ -35,15 +52,15 @@ def add_parents(paths):
 def format_exception(e):
     exc_type, exc_value, exc_tb = sys.exc_info()
     details = traceback.format_exception(exc_type, exc_value, exc_tb)
-    details = '\n'.join(details)
+    details = '\n'.join(map(core.decode, details))
     if hasattr(e, 'msg'):
         msg = e.msg
     else:
-        msg = str(e)
+        msg = core.decode(repr(e))
     return (msg, details)
 
 
-def sublist(a,b):
+def sublist(a, b):
     """Subtracts list b from list a and returns the resulting list."""
     # conceptually, c = a - b
     c = []
@@ -54,11 +71,16 @@ def sublist(a,b):
 
 
 __grep_cache = {}
+
+
 def grep(pattern, items, squash=True):
-    """Greps a list for items that match a pattern and return a list of
-    matching items.  If only one item matches, return just that item.
+    """Greps a list for items that match a pattern
+
+    :param squash: If only one item matches, return just that item
+    :returns: List of matching items
+
     """
-    isdict = type(items) is dict
+    isdict = isinstance(items, dict)
     if pattern in __grep_cache:
         regex = __grep_cache[pattern]
     else:
@@ -83,12 +105,13 @@ def grep(pattern, items, squash=True):
             matched.append(subitems)
 
     if isdict:
-        return matchdict
+        result = matchdict
+    elif squash and len(matched) == 1:
+        result = matched[0]
     else:
-        if squash and len(matched) == 1:
-            return matched[0]
-        else:
-            return matched
+        result = matched
+
+    return result
 
 
 def basename(path):
@@ -103,25 +126,11 @@ def basename(path):
 
 
 def strip_one(path):
-    """Strip one level of directory
-
-    >>> strip_one('/usr/bin/git')
-    u'bin/git'
-
-    >>> strip_one('local/bin/git')
-    u'bin/git'
-
-    >>> strip_one('bin/git')
-    u'git'
-
-    >>> strip_one('git')
-    u'git'
-
-    """
+    """Strip one level of directory"""
     return path.strip('/').split('/', 1)[-1]
 
 
-def dirname(path):
+def dirname(path, current_dir=''):
     """
     An os.path.dirname() implementation that always uses '/'
 
@@ -133,10 +142,53 @@ def dirname(path):
         path = path.replace('//', '/')
     path_dirname = path.rsplit('/', 1)[0]
     if path_dirname == path:
-        return ''
+        return current_dir
     return path.rsplit('/', 1)[0]
 
 
+def splitpath(path):
+    """Split paths using '/' regardless of platform
+
+    """
+    return path.split('/')
+
+
+def join(*paths):
+    """Join paths using '/' regardless of platform
+
+    """
+    return '/'.join(paths)
+
+
+def pathset(path):
+    """Return all of the path components for the specified path
+
+    >>> pathset('foo/bar/baz') == ['foo', 'foo/bar', 'foo/bar/baz']
+    True
+
+    """
+    result = []
+    parts = splitpath(path)
+    prefix = ''
+    for part in parts:
+        result.append(prefix + part)
+        prefix += part + '/'
+
+    return result
+
+
+def select_directory(paths):
+    """Return the first directory in a list of paths"""
+    if not paths:
+        return core.getcwd()
+
+    for path in paths:
+        if core.isdir(path):
+            return path
+
+    return os.path.dirname(paths[0])
+
+
 def strip_prefix(prefix, string):
     """Return string, without the prefix. Blow up if string doesn't
     start with prefix."""
@@ -178,18 +230,13 @@ else:
         return [core.decode(arg) for arg in _shell_split(s)]
 
 
-def tmp_filename(label):
-    label = label.replace('/', '-').replace('\\', '-')
-    fd = tempfile.NamedTemporaryFile(dir=tmpdir(), prefix=label+'-')
+def tmp_filename(label, suffix=''):
+    label = 'git-cola-' + label.replace('/', '-').replace('\\', '-')
+    fd = tempfile.NamedTemporaryFile(prefix=label+'-', suffix=suffix)
     fd.close()
     return fd.name
 
 
-@memoize
-def tmpdir():
-    return tempfile.mkdtemp(prefix='git-cola-')
-
-
 def is_linux():
     """Is this a linux machine?"""
     return sys.platform.startswith('linux')
@@ -244,3 +291,70 @@ class Proxy(object):
 
     def __getattr__(self, name):
         return getattr(self._obj, name)
+
+
+def slice_fn(input_items, map_fn):
+    """Slice input_items and call map_fn over every slice
+
+    This exists because of "errno: Argument list too long"
+
+    """
+    # This comment appeared near the top of include/linux/binfmts.h
+    # in the Linux source tree:
+    #
+    # /*
+    #  * MAX_ARG_PAGES defines the number of pages allocated for arguments
+    #  * and envelope for the new program. 32 should suffice, this gives
+    #  * a maximum env+arg of 128kB w/4KB pages!
+    #  */
+    # #define MAX_ARG_PAGES 32
+    #
+    # 'size' is a heuristic to keep things highly performant by minimizing
+    # the number of slices.  If we wanted it to run as few commands as
+    # possible we could call "getconf ARG_MAX" and make a better guess,
+    # but it's probably not worth the complexity (and the extra call to
+    # getconf that we can't do on Windows anyways).
+    #
+    # In my testing, getconf ARG_MAX on Mac OS X Mountain Lion reported
+    # 262144 and Debian/Linux-x86_64 reported 2097152.
+    #
+    # The hard-coded max_arg_len value is safely below both of these
+    # real-world values.
+
+    # 4K pages x 32 MAX_ARG_PAGES
+    max_arg_len = (32 * 4096) // 4  # allow plenty of space for the environment
+    max_filename_len = 256
+    size = max_arg_len // max_filename_len
+
+    status = 0
+    outs = []
+    errs = []
+
+    items = copy.copy(input_items)
+    while items:
+        stat, out, err = map_fn(items[:size])
+        if stat < 0:
+            status = min(stat, status)
+        else:
+            status = max(stat, status)
+        outs.append(out)
+        errs.append(err)
+        items = items[size:]
+
+    return (status, '\n'.join(outs), '\n'.join(errs))
+
+
+class seq(object):
+
+    def __init__(self, sequence):
+        self.seq = sequence
+
+    def index(self, item, default=-1):
+        try:
+            idx = self.seq.index(item)
+        except ValueError:
+            idx = default
+        return idx
+
+    def __getitem__(self, idx):
+        return self.seq[idx]
