@@ -1,34 +1,44 @@
 from __future__ import division, absolute_import, unicode_literals
+from functools import partial
 
-from PyQt4 import QtGui
-from PyQt4 import QtCore
-from PyQt4.QtCore import Qt
-from PyQt4.QtCore import SIGNAL
+from qtpy import QtCore
+from qtpy import QtGui
+from qtpy import QtWidgets
+from qtpy.QtCore import Qt
+from qtpy.QtCore import Signal
 
-from cola import cmds
-from cola import core
-from cola import gitcmds
-from cola import gitcfg
-from cola import textwrap
-from cola import qtutils
-from cola.cmds import Interaction
-from cola.gitcmds import commit_message_path
-from cola.i18n import N_
-from cola.models import dag
-from cola.models import prefs
-from cola.widgets import defs
-from cola.widgets.selectcommits import select_commits
-from cola.widgets.spellcheck import SpellCheckTextEdit
-from cola.widgets.text import HintedLineEdit
-from cola.compat import ustr
+from .. import actions
+from .. import cmds
+from .. import core
+from .. import gitcmds
+from .. import hotkeys
+from .. import icons
+from .. import textwrap
+from .. import qtutils
+from ..interaction import Interaction
+from ..gitcmds import commit_message_path
+from ..i18n import N_
+from ..models import dag
+from ..models import prefs
+from ..qtutils import get
+from ..utils import Group
+from . import defs
+from .selectcommits import select_commits
+from .spellcheck import SpellCheckTextEdit
+from .text import HintedLineEdit
 
 
-class CommitMessageEditor(QtGui.QWidget):
-    def __init__(self, model, parent):
-        QtGui.QWidget.__init__(self, parent)
+class CommitMessageEditor(QtWidgets.QWidget):
+    commit_message_changed = Signal(object)
+    cursor_changed = Signal(int, int)
+    down = Signal()
+    up = Signal()
+    updated = Signal()
 
-        self.model = model
-        self.notifying = False
+    def __init__(self, context, parent):
+        QtWidgets.QWidget.__init__(self, parent)
+        self.context = context
+        self.model = model = context.model
         self.spellcheck_initialized = False
 
         self._linebreak = None
@@ -36,38 +46,64 @@ class CommitMessageEditor(QtGui.QWidget):
         self._tabwidth = None
 
         # Actions
-        self.signoff_action = qtutils.add_action(self, cmds.SignOff.name(),
-                                                 cmds.run(cmds.SignOff),
-                                                 cmds.SignOff.SHORTCUT)
+        self.signoff_action = qtutils.add_action(
+            self, cmds.SignOff.name(), cmds.run(cmds.SignOff, context),
+            hotkeys.SIGNOFF)
         self.signoff_action.setToolTip(N_('Sign off on this commit'))
 
-        self.commit_action = qtutils.add_action(self,
-                                                N_('Commit@@verb'),
-                                                self.commit,
-                                                cmds.Commit.SHORTCUT)
+        self.commit_action = qtutils.add_action(
+            self, N_('Commit@@verb'), self.commit, hotkeys.COMMIT)
+        self.commit_action.setIcon(icons.commit())
         self.commit_action.setToolTip(N_('Commit staged changes'))
+        self.clear_action = qtutils.add_action(
+            self, N_('Clear...'), self.clear)
+
+        self.launch_editor = actions.launch_editor(context, self)
+        self.launch_difftool = actions.launch_difftool(context, self)
+        self.stage_or_unstage = actions.stage_or_unstage(context, self)
+
+        self.move_up = actions.move_up(self)
+        self.move_down = actions.move_down(self)
+
+        # Menu acctions
+        self.menu_actions = menu_actions = [
+            None,
+            self.signoff_action,
+            self.commit_action,
+            None,
+            self.launch_editor,
+            self.launch_difftool,
+            self.stage_or_unstage,
+            None,
+            self.move_up,
+            self.move_down,
+        ]
 
         # Widgets
-        self.summary = CommitSummaryLineEdit()
+        self.summary = CommitSummaryLineEdit(context)
         self.summary.setMinimumHeight(defs.tool_button_height)
-        self.summary.extra_actions.append(self.signoff_action)
-        self.summary.extra_actions.append(self.commit_action)
+        self.summary.menu_actions.extend(menu_actions)
 
-        self.description = CommitMessageTextEdit()
-        self.description.extra_actions.append(self.signoff_action)
-        self.description.extra_actions.append(self.commit_action)
+        cfg = context.cfg
+        self.summary_validator = MessageValidator(context, parent=self.summary)
+        self.summary.setValidator(self.summary_validator)
+
+        self.description = CommitMessageTextEdit(context, parent=self)
+        self.description.set_dictionary(cfg.get('cola.dictionary', None))
+        self.description.menu_actions.extend(menu_actions)
 
         commit_button_tooltip = N_('Commit staged changes\n'
                                    'Shortcut: Ctrl+Enter')
         self.commit_button = qtutils.create_toolbutton(
-                text=N_('Commit@@verb'), tooltip=commit_button_tooltip,
-                icon=qtutils.save_icon())
+            text=N_('Commit@@verb'), tooltip=commit_button_tooltip,
+            icon=icons.commit())
+        self.commit_group = Group(self.commit_action, self.commit_button)
 
-        self.actions_menu = QtGui.QMenu()
+        self.actions_menu = qtutils.create_menu(N_('Actions'), self)
         self.actions_button = qtutils.create_toolbutton(
-                icon=qtutils.options_icon(), tooltip=N_('Actions...'))
+            icon=icons.configure(), tooltip=N_('Actions...'))
         self.actions_button.setMenu(self.actions_menu)
-        self.actions_button.setPopupMode(QtGui.QToolButton.InstantPopup)
+        self.actions_button.setPopupMode(QtWidgets.QToolButton.InstantPopup)
         qtutils.hide_button_menu_indicator(self.actions_button)
 
         self.actions_menu.addAction(self.signoff_action)
@@ -76,41 +112,47 @@ class CommitMessageEditor(QtGui.QWidget):
 
         # Amend checkbox
         self.amend_action = self.actions_menu.addAction(
-                N_('Amend Last Commit'))
+            N_('Amend Last Commit'))
         self.amend_action.setCheckable(True)
-        self.amend_action.setShortcut(cmds.AmendMode.SHORTCUT)
+        self.amend_action.setShortcut(hotkeys.AMEND)
         self.amend_action.setShortcutContext(Qt.ApplicationShortcut)
 
+        # Bypass hooks
+        self.bypass_commit_hooks_action = self.actions_menu.addAction(
+            N_('Bypass Commit Hooks'))
+        self.bypass_commit_hooks_action.setCheckable(True)
+        self.bypass_commit_hooks_action.setChecked(False)
+
         # Sign commits
-        cfg = gitcfg.current()
         self.sign_action = self.actions_menu.addAction(
-                N_('Create Signed Commit'))
+            N_('Create Signed Commit'))
         self.sign_action.setCheckable(True)
-        self.sign_action.setChecked(cfg.get('cola.signcommits', False))
+        signcommits = cfg.get('cola.signcommits', default=False)
+        self.sign_action.setChecked(signcommits)
 
         # Spell checker
         self.check_spelling_action = self.actions_menu.addAction(
-                N_('Check Spelling'))
+            N_('Check Spelling'))
         self.check_spelling_action.setCheckable(True)
-        self.check_spelling_action.setChecked(False)
+        spellcheck = prefs.spellcheck(context)
+        self.check_spelling_action.setChecked(spellcheck)
+        self.toggle_check_spelling(spellcheck)
 
         # Line wrapping
         self.autowrap_action = self.actions_menu.addAction(
-                N_('Auto-Wrap Lines'))
+            N_('Auto-Wrap Lines'))
         self.autowrap_action.setCheckable(True)
-        self.autowrap_action.setChecked(prefs.linebreak())
+        self.autowrap_action.setChecked(prefs.linebreak(context))
 
         # Commit message
         self.actions_menu.addSeparator()
         self.load_commitmsg_menu = self.actions_menu.addMenu(
-                N_('Load Previous Commit Message'))
-        self.connect(self.load_commitmsg_menu, SIGNAL('aboutToShow()'),
-                     self.build_commitmsg_menu)
+            N_('Load Previous Commit Message'))
+        self.load_commitmsg_menu.aboutToShow.connect(self.build_commitmsg_menu)
 
         self.fixup_commit_menu = self.actions_menu.addMenu(
-                N_('Fixup Previous Commit'))
-        self.connect(self.fixup_commit_menu, SIGNAL('aboutToShow()'),
-                     self.build_fixup_menu)
+            N_('Fixup Previous Commit'))
+        self.fixup_commit_menu.aboutToShow.connect(self.build_fixup_menu)
 
         self.toplayout = qtutils.hbox(defs.no_margin, defs.spacing,
                                       self.actions_button, self.summary,
@@ -125,7 +167,8 @@ class CommitMessageEditor(QtGui.QWidget):
         qtutils.connect_button(self.commit_button, self.commit)
 
         # Broadcast the amend mode
-        qtutils.connect_action_bool(self.amend_action, cmds.run(cmds.AmendMode))
+        qtutils.connect_action_bool(
+            self.amend_action, partial(cmds.run(cmds.AmendMode), context))
         qtutils.connect_action_bool(self.check_spelling_action,
                                     self.toggle_check_spelling)
 
@@ -133,56 +176,68 @@ class CommitMessageEditor(QtGui.QWidget):
         qtutils.connect_action_bool(self.autowrap_action, self.set_linebreak)
 
         qtutils.add_action(self.summary, N_('Move Down'),
-                           self.focus_description,
-                           Qt.Key_Return, Qt.Key_Enter)
+                           self.focus_description, *hotkeys.ACCEPT)
 
         qtutils.add_action(self.summary, N_('Move Down'),
-                           self.summary_cursor_down,
-                           Qt.Key_Down)
+                           self.summary_cursor_down, hotkeys.DOWN)
+
+        self.selection = selection = context.selection
+        selection.add_observer(
+            selection.message_selection_changed, self.updated.emit)
 
         self.model.add_observer(self.model.message_commit_message_changed,
-                                self.set_commit_message)
+                                self.commit_message_changed.emit)
 
-        self.connect(self.summary, SIGNAL('cursorPosition(int,int)'),
-                     self.emit_position)
+        self.commit_message_changed.connect(self.set_commit_message,
+                                            type=Qt.QueuedConnection)
 
-        self.connect(self.description, SIGNAL('cursorPosition(int,int)'),
-                     # description starts at line 2
-                     lambda row, col: self.emit_position(row + 2, col))
+        self.summary.cursor_changed.connect(self.cursor_changed.emit)
+        self.description.cursor_changed.connect(
+            # description starts at line 2
+            lambda row, col: self.cursor_changed.emit(row + 2, col))
 
-        # Keep model informed of changes
-        self.connect(self.summary, SIGNAL('textChanged(QString)'),
-                     self.commit_summary_changed)
+        self.summary.textChanged.connect(self.commit_summary_changed)
+        self.description.textChanged.connect(self._commit_message_changed)
+        self.description.leave.connect(self.focus_summary)
+        self.updated.connect(self.refresh)
 
-        self.connect(self.description, SIGNAL('textChanged()'),
-                     self.commit_message_changed)
+        self.commit_group.setEnabled(False)
 
-        self.connect(self.description, SIGNAL('leave()'),
-                     self.focus_summary)
-
-        self.setFont(qtutils.diff_font())
-
-        self.summary.enable_hint(True)
-        self.description.enable_hint(True)
-
-        self.commit_button.setEnabled(False)
-        self.commit_action.setEnabled(False)
-
-        self.setFocusProxy(self.summary)
-
-        self.set_tabwidth(prefs.tabwidth())
-        self.set_textwidth(prefs.textwidth())
-        self.set_linebreak(prefs.linebreak())
+        self.set_expandtab(prefs.expandtab(context))
+        self.set_tabwidth(prefs.tabwidth(context))
+        self.set_textwidth(prefs.textwidth(context))
+        self.set_linebreak(prefs.linebreak(context))
 
         # Loading message
         commit_msg = ''
-        commit_msg_path = commit_message_path()
+        commit_msg_path = commit_message_path(context)
         if commit_msg_path:
             commit_msg = core.read(commit_msg_path)
-        self.set_commit_message(commit_msg)
+        model.set_commitmsg(commit_msg)
 
         # Allow tab to jump from the summary to the description
         self.setTabOrder(self.summary, self.description)
+        self.setFont(qtutils.diff_font(context))
+        self.setFocusProxy(self.summary)
+
+        cfg.add_observer(cfg.message_user_config_changed, self.config_changed)
+
+    def config_changed(self, key, value):
+        if key != prefs.SPELL_CHECK:
+            return
+        if get(self.check_spelling_action) == value:
+            return
+        self.check_spelling_action.setChecked(value)
+        self.toggle_check_spelling(value)
+
+    def refresh(self):
+        enabled = self.model.stageable() or self.model.unstageable()
+        if self.model.stageable():
+            text = N_('Stage')
+        else:
+            text = N_('Unstage')
+        self.stage_or_unstage.setEnabled(enabled)
+        self.stage_or_unstage.setText(text)
 
     def set_initial_size(self):
         self.setMaximumHeight(133)
@@ -206,7 +261,7 @@ class CommitMessageEditor(QtGui.QWidget):
 
         """
         cur_position = self.summary.cursorPosition()
-        end_position = len(self.summary.value())
+        end_position = len(get(self.summary))
         if cur_position == end_position:
             self.focus_description()
         else:
@@ -214,22 +269,21 @@ class CommitMessageEditor(QtGui.QWidget):
 
     def commit_message(self, raw=True):
         """Return the commit message as a unicode string"""
-        summary = self.summary.value()
+        summary = get(self.summary)
         if raw:
-            description = self.description.value()
+            description = get(self.description)
         else:
             description = self.formatted_description()
         if summary and description:
             return summary + '\n\n' + description
-        elif summary:
+        if summary:
             return summary
-        elif description:
+        if description:
             return '\n\n' + description
-        else:
-            return ''
+        return ''
 
     def formatted_description(self):
-        text = self.description.value()
+        text = get(self.description)
         if not self._linebreak:
             return text
         return textwrap.word_wrap(text, self._tabwidth, self._textwidth)
@@ -243,50 +297,47 @@ class CommitMessageEditor(QtGui.QWidget):
         "extended description" field.
 
         """
-        value = ustr(value)
         if '\n' in value:
             summary, description = value.split('\n', 1)
             description = description.lstrip('\n')
-            cur_description = self.description.value()
+            cur_description = get(self.description)
             if cur_description:
                 description = description + '\n' + cur_description
             # this callback is triggered by changing `summary`
             # so disable signals for `summary` only.
-            self.summary.blockSignals(True)
-            self.summary.set_value(summary)
-            self.summary.blockSignals(False)
+            self.summary.set_value(summary, block=True)
             self.description.set_value(description)
-        self.commit_message_changed()
+        self._commit_message_changed()
 
-    def commit_message_changed(self, value=None):
+    def _commit_message_changed(self, _value=None):
         """Update the model when values change"""
-        self.notifying = True
         message = self.commit_message()
-        self.model.set_commitmsg(message)
+        self.model.set_commitmsg(message, notify=False)
         self.refresh_palettes()
-        self.notifying = False
         self.update_actions()
 
+    def clear(self):
+        if not Interaction.confirm(
+                N_('Clear commit message?'),
+                N_('The commit message will be cleared.'),
+                N_('This cannot be undone.  Clear commit message?'),
+                N_('Clear commit message'), default=True, icon=icons.discard()):
+            return
+        self.model.set_commitmsg('')
+
     def update_actions(self):
-        commit_enabled = bool(self.summary.value())
-        self.commit_button.setEnabled(commit_enabled)
-        self.commit_action.setEnabled(commit_enabled)
+        commit_enabled = bool(get(self.summary))
+        self.commit_group.setEnabled(commit_enabled)
 
     def refresh_palettes(self):
         """Update the color palette for the hint text"""
-        self.summary.refresh_palette()
-        self.description.refresh_palette()
+        self.summary.hint.refresh()
+        self.description.hint.refresh()
 
     def set_commit_message(self, message):
         """Set the commit message to match the observed model"""
-        if self.notifying:
-            # Calling self.model.set_commitmsg(message) causes us to
-            # loop around so break the loop
-            return
-
         # Parse the "summary" and "description" fields
-        umsg = ustr(message)
-        lines = umsg.splitlines()
+        lines = message.splitlines()
 
         num_lines = len(lines)
 
@@ -319,21 +370,10 @@ class CommitMessageEditor(QtGui.QWidget):
         focus_description = not description
 
         # Update summary
-        if not summary and not self.summary.hasFocus():
-            summary = self.summary.hint()
-
-        blocksignals = self.summary.blockSignals(True)
-        self.summary.setText(summary)
-        self.summary.setCursorPosition(0)
-        self.summary.blockSignals(blocksignals)
+        self.summary.set_value(summary, block=True)
 
         # Update description
-        if not description and not self.description.hasFocus():
-            description = self.description.hint()
-
-        blocksignals = self.description.blockSignals(True)
-        self.description.setPlainText(description)
-        self.description.blockSignals(blocksignals)
+        self.description.set_value(description, block=True)
 
         # Update text color
         self.refresh_palettes()
@@ -341,15 +381,16 @@ class CommitMessageEditor(QtGui.QWidget):
         # Focus the empty summary or description
         if focus_summary:
             self.summary.setFocus()
-            self.summary.emit_position()
         elif focus_description:
             self.description.setFocus()
-            self.description.emit_position()
         else:
-            self.summary.emit_position()
+            self.summary.cursor_position.emit()
 
         self.update_actions()
 
+    def set_expandtab(self, value):
+        self.description.set_expandtab(value)
+
     def set_tabwidth(self, width):
         self._tabwidth = width
         self.description.set_tabwidth(width)
@@ -378,14 +419,12 @@ class CommitMessageEditor(QtGui.QWidget):
         self.amend_action.setChecked(checked)
         self.amend_action.blockSignals(blocksignals)
 
-    def emit_position(self, row, col):
-        self.emit(SIGNAL('cursorPosition(int,int)'), row, col)
-
     def commit(self):
         """Attempt to create a commit from the index and commit message."""
-        if not bool(self.summary.value()):
+        context = self.context
+        if not bool(get(self.summary)):
             # Describe a good commit message
-            error_msg = N_(''
+            error_msg = N_(
                 'Please supply a commit message.\n\n'
                 'A good commit message has the following format:\n\n'
                 '- First line: Describe in one sentence what you did.\n'
@@ -397,44 +436,41 @@ class CommitMessageEditor(QtGui.QWidget):
 
         msg = self.commit_message(raw=False)
 
-        if not self.model.staged:
-            error_msg = N_(''
+        # We either need to have something staged, or be merging.
+        # If there was a merge conflict resolved, there may not be anything
+        # to stage, but we still need to commit to complete the merge.
+        if not (self.model.staged or self.model.is_merging):
+            error_msg = N_(
                 'No changes to commit.\n\n'
                 'You must stage at least 1 file before you can commit.')
             if self.model.modified:
                 informative_text = N_('Would you like to stage and '
                                       'commit all modified files?')
-                if not qtutils.confirm(
-                        N_('Stage and commit?'),
-                        error_msg, informative_text,
+                if not Interaction.confirm(
+                        N_('Stage and commit?'), error_msg, informative_text,
                         N_('Stage and Commit'),
-                        default=True,
-                        icon=qtutils.save_icon()):
+                        default=True, icon=icons.save()):
                     return
             else:
                 Interaction.information(N_('Nothing to commit'), error_msg)
                 return
-            cmds.do(cmds.StageModified)
+            cmds.do(cmds.StageModified, context)
 
         # Warn that amending published commits is generally bad
-        amend = self.amend_action.isChecked()
+        amend = get(self.amend_action)
         if (amend and self.model.is_commit_published() and
-            not qtutils.confirm(
-                        N_('Rewrite Published Commit?'),
-                        N_('This commit has already been published.\n'
-                           'This operation will rewrite published history.\n'
-                           'You probably don\'t want to do this.'),
-                        N_('Amend the published commit?'),
-                        N_('Amend Commit'),
-                        default=False, icon=qtutils.save_icon())):
+                not Interaction.confirm(
+                    N_('Rewrite Published Commit?'),
+                    N_('This commit has already been published.\n'
+                       'This operation will rewrite published history.\n'
+                       'You probably don\'t want to do this.'),
+                    N_('Amend the published commit?'),
+                    N_('Amend Commit'), default=False, icon=icons.save())):
             return
-        sign = self.sign_action.isChecked()
-        status, out, err = cmds.do(cmds.Commit, amend, msg, sign)
-        if status != 0:
-            Interaction.critical(N_('Commit failed'),
-                                 N_('"git commit" returned exit code %s') %
-                                    (status,),
-                                 out + err)
+        no_verify = get(self.bypass_commit_hooks_action)
+        sign = get(self.sign_action)
+        cmds.do(cmds.Commit, context, amend, msg, sign, no_verify=no_verify)
+        self.bypass_commit_hooks_action.setChecked(False)
 
     def build_fixup_menu(self):
         self.build_commits_menu(cmds.LoadFixupMessage,
@@ -443,51 +479,54 @@ class CommitMessageEditor(QtGui.QWidget):
                                 prefix='fixup! ')
 
     def build_commitmsg_menu(self):
-        self.build_commits_menu(cmds.LoadCommitMessageFromSHA1,
+        self.build_commits_menu(cmds.LoadCommitMessageFromOID,
                                 self.load_commitmsg_menu,
                                 self.choose_commit_message)
 
     def build_commits_menu(self, cmd, menu, chooser, prefix=''):
-        ctx = dag.DAG('HEAD', 6)
-        commits = dag.RepoReader(ctx)
+        context = self.context
+        params = dag.DAG('HEAD', 6)
+        commits = dag.RepoReader(context, params)
 
         menu_commits = []
-        for idx, c in enumerate(commits):
+        for idx, c in enumerate(commits.get()):
             menu_commits.insert(0, c)
             if idx > 5:
                 continue
 
         menu.clear()
         for c in menu_commits:
-            menu.addAction(prefix + c.summary, cmds.run(cmd, c.sha1))
+            menu.addAction(prefix + c.summary, cmds.run(cmd, context, c.oid))
 
         if len(commits) == 6:
             menu.addSeparator()
             menu.addAction(N_('More...'), chooser)
 
-
     def choose_commit(self, cmd):
-        revs, summaries = gitcmds.log_helper()
-        sha1s = select_commits(N_('Select Commit'), revs, summaries,
-                               multiselect=False)
-        if not sha1s:
+        context = self.context
+        revs, summaries = gitcmds.log_helper(context)
+        oids = select_commits(
+            context, N_('Select Commit'), revs, summaries, multiselect=False)
+        if not oids:
             return
-        sha1 = sha1s[0]
-        cmds.do(cmd, sha1)
+        oid = oids[0]
+        cmds.do(cmd, context, oid)
 
     def choose_commit_message(self):
-        self.choose_commit(cmds.LoadCommitMessageFromSHA1)
+        self.choose_commit(cmds.LoadCommitMessageFromOID)
 
     def choose_fixup_commit(self):
         self.choose_commit(cmds.LoadFixupMessage)
 
     def toggle_check_spelling(self, enabled):
         spellcheck = self.description.spellcheck
+        cfg = self.context.cfg
 
+        if cfg.get_user(prefs.SPELL_CHECK) != enabled:
+            cfg.set_user(prefs.SPELL_CHECK, enabled)
         if enabled and not self.spellcheck_initialized:
             # Add our name to the dictionary
             self.spellcheck_initialized = True
-            cfg = gitcfg.current()
             user_name = cfg.get('user.name')
             if user_name:
                 for part in user_name.split():
@@ -509,41 +548,78 @@ class CommitMessageEditor(QtGui.QWidget):
         self.description.highlighter.enable(enabled)
 
 
-class CommitSummaryLineEdit(HintedLineEdit):
-    def __init__(self, parent=None):
-        hint = N_('Commit summary')
-        HintedLineEdit.__init__(self, hint, parent)
-        self.extra_actions = []
+class MessageValidator(QtGui.QValidator):
+    """Prevent invalid branch names"""
 
-        regex = QtCore.QRegExp(r'^[^#].*')
-        self._validator = QtGui.QRegExpValidator(regex, self)
-        self.setValidator(self._validator)
+    config_updated = Signal()
+
+    def __init__(self, context, parent=None):
+        super(MessageValidator, self).__init__(parent)
+        self.context = context
+        self._comment_char = None
+        self._cfg = cfg = context.cfg
+        self.refresh()
+        self.config_updated.connect(self.refresh, type=Qt.QueuedConnection)
+        cfg.add_observer(cfg.message_updated, self.emit_config_updated)
+        self.destroyed.connect(self.teardown)
+
+    def teardown(self):
+        self._cfg.remove_observer(self.emit_config_updated)
+
+    def emit_config_updated(self):
+        self.config_updated.emit()
+
+    def refresh(self):
+        """Update comment char in response to config changes"""
+        self._comment_char = prefs.comment_char(self.context)
+
+    def validate(self, string, idx):
+        """Scrub whitespace and validate the commit message"""
+        string = string.lstrip()
+        if string.startswith(self._comment_char):
+            state = self.Invalid
+        else:
+            state = self.Acceptable
+        return (state, string, idx)
+
+
+class CommitSummaryLineEdit(HintedLineEdit):
+
+    cursor = Signal(int, int)
+
+    def __init__(self, context, parent=None):
+        hint = N_('Commit summary')
+        HintedLineEdit.__init__(self, context, hint, parent=parent)
+        self.menu_actions = []
+
+    def build_menu(self):
+        menu = self.createStandardContextMenu()
+        add_menu_actions(menu, self.menu_actions)
+        return menu
 
     def contextMenuEvent(self, event):
-        menu = self.createStandardContextMenu()
-        if self.extra_actions:
-            menu.addSeparator()
-        for action in self.extra_actions:
-            menu.addAction(action)
+        menu = self.build_menu()
         menu.exec_(self.mapToGlobal(event.pos()))
 
 
 class CommitMessageTextEdit(SpellCheckTextEdit):
+    leave = Signal()
 
-    def __init__(self, parent=None):
+    def __init__(self, context, parent=None):
         hint = N_('Extended description...')
-        SpellCheckTextEdit.__init__(self, hint, parent)
-        self.extra_actions = []
+        SpellCheckTextEdit.__init__(self, context, hint, parent)
+        self.menu_actions = []
 
-        self.action_emit_leave = qtutils.add_action(self,
-                'Shift Tab', self.emit_leave, 'Shift+tab')
+        self.action_emit_leave = qtutils.add_action(
+            self, 'Shift Tab', self.leave.emit, hotkeys.LEAVE)
+
+    def build_menu(self):
+        menu, _ = self.context_menu()
+        add_menu_actions(menu, self.menu_actions)
+        return menu
 
     def contextMenuEvent(self, event):
-        menu, spell_menu = self.context_menu()
-        if self.extra_actions:
-            menu.addSeparator()
-        for action in self.extra_actions:
-            menu.addAction(action)
+        menu = self.build_menu()
         menu.exec_(self.mapToGlobal(event.pos()))
 
     def keyPressEvent(self, event):
@@ -559,10 +635,10 @@ class CommitMessageTextEdit(SpellCheckTextEdit):
                     cursor.setPosition(0)
                     self.setTextCursor(cursor)
                 else:
-                    self.emit_leave()
+                    self.leave.emit()
                 event.accept()
                 return
-            text_before = ustr(self.toPlainText())[:position]
+            text_before = self.toPlainText()[:position]
             lines_before = text_before.count('\n')
             if lines_before == 0:
                 # If we're on the first line, but not at the
@@ -579,7 +655,7 @@ class CommitMessageTextEdit(SpellCheckTextEdit):
         elif event.key() == Qt.Key_Down:
             cursor = self.textCursor()
             position = cursor.position()
-            all_text = ustr(self.toPlainText())
+            all_text = self.toPlainText()
             text_after = all_text[position:]
             lines_after = text_after.count('\n')
             if lines_after == 0:
@@ -593,10 +669,16 @@ class CommitMessageTextEdit(SpellCheckTextEdit):
                 return
         SpellCheckTextEdit.keyPressEvent(self, event)
 
-    def emit_leave(self):
-        self.emit(SIGNAL('leave()'))
-
     def setFont(self, font):
         SpellCheckTextEdit.setFont(self, font)
         fm = self.fontMetrics()
         self.setMinimumSize(QtCore.QSize(1, fm.height() * 2))
+
+
+def add_menu_actions(menu, menu_actions):
+    """Add actions to a menu, treating None as a separator"""
+    for action in menu_actions:
+        if action is None:
+            menu.addSeparator()
+        else:
+            menu.addAction(action)
