@@ -1,49 +1,101 @@
 from __future__ import division, absolute_import, unicode_literals
-
 import re
-import subprocess
 
-from PyQt4 import QtCore
-from PyQt4 import QtGui
-from PyQt4.QtCore import Qt
-from PyQt4.QtCore import SIGNAL
+from qtpy import QtCore
+from qtpy import QtGui
+from qtpy import QtWidgets
+from qtpy.QtCore import Qt
+from qtpy.QtCore import Signal
 
-from cola import gitcmds
-from cola import icons
-from cola import qtutils
-from cola import utils
-from cola.models import main
-from cola.widgets import defs
-from cola.widgets import text
-from cola.compat import ustr
+from .. import core
+from .. import gitcmds
+from .. import icons
+from .. import qtutils
+from .. import utils
+from . import defs
+from .text import HintedLineEdit
 
 
-UPDATE_SIGNAL = 'update()'
+class ValidateRegex(object):
+
+    def __init__(self, regex):
+        self.regex = re.compile(regex)  # regex to scrub
+
+    def validate(self, string, idx):
+        """Scrub and validate the user-supplied input"""
+        state = QtGui.QValidator.Acceptable
+        if self.regex.search(string):
+            string = self.regex.sub('', string)  # scrub matching bits
+            idx = min(idx-1, len(string))
+        return (state, string, idx)
 
 
-class CompletionLineEdit(text.HintedLineEdit):
-    """An lineedit with advanced completion abilities"""
+class RemoteValidator(QtGui.QValidator):
+    """Prevent invalid remote names"""
+
+    def __init__(self, parent=None):
+        super(RemoteValidator, self).__init__(parent)
+        self._validate = ValidateRegex(r'[ \t\\/]')
+
+    def validate(self, string, idx):
+        return self._validate.validate(string, idx)
+
+
+class BranchValidator(QtGui.QValidator):
+    """Prevent invalid branch names"""
+
+    def __init__(self, git, parent=None):
+        super(BranchValidator, self).__init__(parent)
+        self._git = git
+        self._validate = ValidateRegex(r'[ \t\\]')  # forward-slash is okay
+
+    def validate(self, string, idx):
+        """Scrub and validate the user-supplied input"""
+        state, string, idx = self._validate.validate(string, idx)
+        if string:  # Allow empty strings
+            status, _, _ = self._git.check_ref_format(string, branch=True)
+            if status != 0:
+                # The intermediate string, when deleting characters, might
+                # end in a name that is invalid to Git, but we must allow it
+                # otherwise we won't be able to delete it using backspace.
+                if string.endswith('/') or string.endswith('.'):
+                    state = self.Intermediate
+                else:
+                    state = self.Invalid
+        return (state, string, idx)
+
+
+class CompletionLineEdit(HintedLineEdit):
+    """A lineedit with advanced completion abilities"""
+
+    activated = Signal()
+    changed = Signal()
+    cleared = Signal()
+    enter = Signal()
+    up = Signal()
+    down = Signal()
 
     # Activation keys will cause a selected completion item to be chosen
     ACTIVATION_KEYS = (Qt.Key_Return, Qt.Key_Enter)
 
     # Navigation keys trigger signals that widgets can use for customization
     NAVIGATION_KEYS = {
-            Qt.Key_Return: 'return()',
-            Qt.Key_Enter: 'enter()',
-            Qt.Key_Up: 'up()',
-            Qt.Key_Down: 'down()',
+        Qt.Key_Return: 'enter',
+        Qt.Key_Enter: 'enter',
+        Qt.Key_Up: 'up',
+        Qt.Key_Down: 'down',
     }
 
-    def __init__(self, model_factory, hint='', parent=None):
-        text.HintedLineEdit.__init__(self, hint=hint, parent=parent)
+    def __init__(self, context, model_factory, hint='', parent=None):
+        HintedLineEdit.__init__(self, context, hint, parent=parent)
         # Tracks when the completion popup was active during key events
-        self._was_visible = False
+
+        self.context = context
         # The most recently selected completion item
         self._selection = None
 
         # Create a completion model
-        completion_model = model_factory(self)
+        completion_model = model_factory(context, self)
         completer = Completer(completion_model, self)
         completer.setWidget(self)
         self._completer = completer
@@ -53,26 +105,19 @@ class CompletionLineEdit(text.HintedLineEdit):
         self._delegate = HighlightDelegate(self)
         completer.popup().setItemDelegate(self._delegate)
 
-        self.connect(self, SIGNAL('textChanged(QString)'), self._text_changed)
-
-        self.connect(self._completer, SIGNAL('activated(QString)'),
-                     self.choose_completion)
-
-        self.connect(self._completion_model, SIGNAL('updated()'),
-                     self._completions_updated, Qt.QueuedConnection)
-
-        self.connect(self, SIGNAL('destroyed(QObject*)'), self.dispose)
+        self.textChanged.connect(self._text_changed)
+        self._completer.activated.connect(self.choose_completion)
+        self._completion_model.updated.connect(self._completions_updated,
+                                               type=Qt.QueuedConnection)
+        self.destroyed.connect(self.dispose)
 
     def __del__(self):
         self.dispose()
 
+    # pylint: disable=unused-argument
     def dispose(self, *args):
         self._completer.dispose()
 
-    def was_visible(self):
-        """Was the popup visible during the last keypress event?"""
-        return self._was_visible
-
     def completion_selection(self):
         """Return the last completion's selection"""
         return self._selection
@@ -92,9 +137,8 @@ class CompletionLineEdit(text.HintedLineEdit):
     def _is_case_sensitive(self, text):
         return bool([char for char in text if char.isupper()])
 
-    def _text_changed(self, text):
+    def _text_changed(self, full_text):
         match_text = self._last_word()
-        full_text = ustr(text)
         self._do_text_changed(full_text, match_text)
         self.complete_last_word()
 
@@ -118,7 +162,6 @@ class CompletionLineEdit(text.HintedLineEdit):
         This is the event handler for the QCompleter.activated(QString) signal,
         it is called when the user selects an item in the completer popup.
         """
-        completion = ustr(completion)
         if not completion:
             self._do_text_changed('', '')
             return
@@ -127,9 +170,9 @@ class CompletionLineEdit(text.HintedLineEdit):
             words.pop()
 
         words.append(completion)
-        text = subprocess.list2cmdline(words)
+        text = core.list2cmdline(words)
         self.setText(text)
-        self.emit(SIGNAL('changed()'))
+        self.changed.emit()
         self._do_text_changed(text, '')
         self.popup().hide()
 
@@ -137,7 +180,8 @@ class CompletionLineEdit(text.HintedLineEdit):
         return utils.shell_split(self.value())
 
     def _ends_with_whitespace(self):
-        return self.value() != self.value().rstrip()
+        value = self.value()
+        return value != value.rstrip()
 
     def _last_word(self):
         if self._ends_with_whitespace():
@@ -149,11 +193,6 @@ class CompletionLineEdit(text.HintedLineEdit):
             return ''
         return words[-1]
 
-    def event(self, event):
-        if event.type() == QtCore.QEvent.Hide:
-            self.close_popup()
-        return text.HintedLineEdit.event(self, event)
-
     def complete_last_word(self):
         self.update_matches()
         self.complete()
@@ -162,24 +201,18 @@ class CompletionLineEdit(text.HintedLineEdit):
         if self.popup().isVisible():
             self.popup().close()
 
-    def _update_popup_items(self, prefix):
-        """
-        Filters the completer's popup items to only show items
-        with the given prefix.
-        """
-        self._completer.setCompletionPrefix(prefix)
-
     def _completions_updated(self):
         popup = self.popup()
         if not popup.isVisible():
             return
         # Select the first item
         idx = self._completion_model.index(0, 0)
-        selection = QtGui.QItemSelection(idx, idx)
-        mode = QtGui.QItemSelectionModel.Select
+        selection = QtCore.QItemSelection(idx, idx)
+        mode = QtCore.QItemSelectionModel.Select
         popup.selectionModel().select(selection, mode)
 
     def selected_completion(self):
+        """Return the selected completion item"""
         popup = self.popup()
         if not popup.isVisible():
             return None
@@ -190,103 +223,116 @@ class CompletionLineEdit(text.HintedLineEdit):
         idx = indexes[0]
         item = self._completion_model.itemFromIndex(idx)
         if not item:
-            return
-        return ustr(item.text())
-
-    # Qt events
-    def keyPressEvent(self, event):
-        self._was_visible = visible = self.popup().isVisible()
-        key = event.key()
-        was_empty = not bool(self.value())
+            return None
+        return item.text()
 
+    def select_completion(self):
+        """Choose the selected completion option from the completion popup"""
+        result = False
+        visible = self.popup().isVisible()
         if visible:
-            self._selection = self.selected_completion()
-        else:
-            self._selection = None
-            if event.key() in self.ACTIVATION_KEYS:
-                event.accept()
-                return
+            selection = self.selected_completion()
+            if selection:
+                self.choose_completion(selection)
+                result = True
+        return result
 
-        result = text.HintedLineEdit.keyPressEvent(self, event)
+    # Qt overrides
+    def event(self, event):
+        """Override QWidget::event() for tab completion"""
+        event_type = event.type()
 
-        # Backspace at the beginning of the line should hide the popup
-        if was_empty and visible and key == Qt.Key_Backspace:
-            self.popup().hide()
-        # Clearing a line should always emit a signal
-        is_empty = not bool(self.value())
+        if (event_type == QtCore.QEvent.KeyPress
+                and event.key() == Qt.Key_Tab
+                and self.select_completion()):
+            return True
+
+        # Make sure the popup goes away during teardown
+        if event_type == QtCore.QEvent.Hide:
+            self.close_popup()
+
+        return super(CompletionLineEdit, self).event(event)
+
+    def keyPressEvent(self, event):
+        """Process completion and navigation events"""
+        super(CompletionLineEdit, self).keyPressEvent(event)
+        visible = self.popup().isVisible()
+
+        # Hide the popup when the field is empty
+        is_empty = not self.value()
         if is_empty:
-            self.emit(SIGNAL('cleared()'))
-        return result
+            self.cleared.emit()
+            if visible:
+                self.popup().hide()
 
-    def keyReleaseEvent(self, event):
-        """React to release events, handle completion"""
+        # Activation keys select the completion when pressed and emit the
+        # activated signal.  Navigation keys have lower priority, and only
+        # emit when it wasn't already handled as an activation event.
         key = event.key()
-        visible = self.was_visible()
-        if not visible:
-            # If it's a navigation key then emit a signal
-            try:
-                msg = self.NAVIGATION_KEYS[key]
-                event.accept()
-                self.emit(SIGNAL(msg))
-                return
-            except KeyError:
-                pass
-        # Run the real release event
-        result = text.HintedLineEdit.keyReleaseEvent(self, event)
-        # If the popup was visible and we have a selected popup item
-        # then choose that completion.
-        selection = self.completion_selection()
-        if visible and selection and key in self.ACTIVATION_KEYS:
-            self.choose_completion(selection)
-            self.emit(SIGNAL('activated()'))
+        if key in self.ACTIVATION_KEYS and visible:
+            if self.select_completion():
+                self.activated.emit()
             return
-        return result
+
+        navigation = self.NAVIGATION_KEYS.get(key, None)
+        if navigation:
+            signal = getattr(self, navigation)
+            signal.emit()
 
 
 class GatherCompletionsThread(QtCore.QThread):
 
+    items_gathered = Signal(object)
+
     def __init__(self, model):
         QtCore.QThread.__init__(self)
         self.model = model
         self.case_sensitive = False
+        self.running = False
+
+    def dispose(self):
+        self.running = False
+        self.wait()
 
     def run(self):
         text = None
+        self.running = True
         # Loop when the matched text changes between the start and end time.
         # This happens when gather_matches() takes too long and the
         # model's match_text changes in-between.
-        while text != self.model.match_text:
+        while self.running and text != self.model.match_text:
             text = self.model.match_text
             items = self.model.gather_matches(self.case_sensitive)
 
-        if text is not None:
-            self.emit(SIGNAL('items_gathered(PyQt_PyObject)'), items)
+        if self.running and text is not None:
+            self.items_gathered.emit(items)
 
 
-class HighlightDelegate(QtGui.QStyledItemDelegate):
+class HighlightDelegate(QtWidgets.QStyledItemDelegate):
     """A delegate used for auto-completion to give formatted completion"""
-    def __init__(self, parent=None): # model, parent=None):
-        QtGui.QStyledItemDelegate.__init__(self, parent)
+
+    def __init__(self, parent):
+        QtWidgets.QStyledItemDelegate.__init__(self, parent)
+        self.widget = parent
         self.highlight_text = ''
         self.case_sensitive = False
 
         self.doc = QtGui.QTextDocument()
-        try:
+        # older PyQt4 does not have setDocumentMargin
+        if hasattr(self.doc, 'setDocumentMargin'):
             self.doc.setDocumentMargin(0)
-        except: # older PyQt4
-            pass
 
     def set_highlight_text(self, text, case_sensitive):
-        """Sets the text that will be made bold in the term name when displayed"""
+        """Sets the text that will be made bold when displayed"""
         self.highlight_text = text
         self.case_sensitive = case_sensitive
 
     def paint(self, painter, option, index):
         """Overloaded Qt method for custom painting of a model index"""
         if not self.highlight_text:
-            return QtGui.QStyledItemDelegate.paint(self, painter, option, index)
-
-        text = ustr(index.data().toPyObject())
+            QtWidgets.QStyledItemDelegate.paint(self, painter, option, index)
+            return
+        text = index.data()
         if self.case_sensitive:
             html = text.replace(self.highlight_text,
                                 '<strong>%s</strong>' % self.highlight_text)
@@ -303,25 +349,26 @@ class HighlightDelegate(QtGui.QStyledItemDelegate):
         self.doc.setHtml(html)
 
         # Painting item without text, Text Document will paint the text
-        optionV4 = QtGui.QStyleOptionViewItemV4(option)
-        self.initStyleOption(optionV4, index)
-        optionV4.text = QtCore.QString()
+        params = QtWidgets.QStyleOptionViewItem(option)
+        self.initStyleOption(params, index)
+        params.text = ''
 
-        style = QtGui.QApplication.style()
-        style.drawControl(QtGui.QStyle.CE_ItemViewItem, optionV4, painter)
+        style = QtWidgets.QApplication.style()
+        style.drawControl(QtWidgets.QStyle.CE_ItemViewItem, params, painter)
         ctx = QtGui.QAbstractTextDocumentLayout.PaintContext()
 
         # Highlighting text if item is selected
-        if (optionV4.state & QtGui.QStyle.State_Selected):
-            color = optionV4.palette.color(QtGui.QPalette.Active,
-                                           QtGui.QPalette.HighlightedText)
+        if params.state & QtWidgets.QStyle.State_Selected:
+            color = params.palette.color(QtGui.QPalette.Active,
+                                         QtGui.QPalette.HighlightedText)
             ctx.palette.setColor(QtGui.QPalette.Text, color)
 
         # translate the painter to where the text is drawn
-        rect = style.subElementRect(QtGui.QStyle.SE_ItemViewItemText, optionV4)
+        item_text = QtWidgets.QStyle.SE_ItemViewItemText
+        rect = style.subElementRect(item_text, params, self.widget)
         painter.save()
 
-        start = rect.topLeft() + QtCore.QPoint(3, 0)
+        start = rect.topLeft() + QtCore.QPoint(defs.margin, 0)
         painter.translate(start)
 
         # tell the text document to draw the html for us
@@ -338,16 +385,20 @@ def ref_sort_key(ref):
 
 class CompletionModel(QtGui.QStandardItemModel):
 
-    def __init__(self, parent):
+    updated = Signal()
+    items_gathered = Signal(object)
+    model_updated = Signal()
+
+    def __init__(self, context, parent):
         QtGui.QStandardItemModel.__init__(self, parent)
+        self.context = context
         self.match_text = ''
         self.full_text = ''
         self.case_sensitive = False
 
         self.update_thread = GatherCompletionsThread(self)
-        self.connect(self.update_thread,
-                     SIGNAL('items_gathered(PyQt_PyObject)'),
-                     self.apply_matches, Qt.QueuedConnection)
+        self.update_thread.items_gathered.connect(self.apply_matches,
+                                                  type=Qt.QueuedConnection)
 
     def update(self):
         case_sensitive = self.update_thread.case_sensitive
@@ -364,6 +415,7 @@ class CompletionModel(QtGui.QStandardItemModel):
         if not self.update_thread.isRunning():
             self.update_thread.start()
 
+    # pylint: disable=unused-argument
     def gather_matches(self, case_sensitive):
         return ((), (), set())
 
@@ -391,9 +443,23 @@ class CompletionModel(QtGui.QStandardItemModel):
                 item.setIcon(from_filename(match))
             items.append(item)
 
-        self.clear()
-        self.invisibleRootItem().appendRows(items)
-        self.emit(SIGNAL('updated()'))
+        try:
+            self.clear()
+            self.invisibleRootItem().appendRows(items)
+            self.updated.emit()
+        except RuntimeError:  # C++ object has been deleted
+            pass
+
+    def dispose(self):
+        self.update_thread.dispose()
+
+
+def _identity(x):
+    return x
+
+
+def _lower(x):
+    return x.lower()
 
 
 def filter_matches(match_text, candidates, case_sensitive,
@@ -401,9 +467,9 @@ def filter_matches(match_text, candidates, case_sensitive,
     """Filter candidates and return the matches"""
 
     if case_sensitive:
-        case_transform = lambda x: x
+        case_transform = _identity
     else:
-        case_transform = lambda x: x.lower()
+        case_transform = _lower
 
     if match_text:
         match_text = case_transform(match_text)
@@ -426,16 +492,15 @@ def filter_path_matches(match_text, file_list, case_sensitive):
     return (paths, dirs)
 
 
-class Completer(QtGui.QCompleter):
+class Completer(QtWidgets.QCompleter):
 
     def __init__(self, model, parent):
-        QtGui.QCompleter.__init__(self, parent)
+        QtWidgets.QCompleter.__init__(self, parent)
         self._model = model
-        self.setCompletionMode(QtGui.QCompleter.UnfilteredPopupCompletion)
+        self.setCompletionMode(QtWidgets.QCompleter.UnfilteredPopupCompletion)
         self.setCaseSensitivity(Qt.CaseInsensitive)
 
-        self.connect(model, SIGNAL(UPDATE_SIGNAL),
-                     self.update, Qt.QueuedConnection)
+        model.model_updated.connect(self.update, type=Qt.QueuedConnection)
         self.setModel(model)
 
     def update(self):
@@ -450,66 +515,79 @@ class Completer(QtGui.QCompleter):
 
 class GitCompletionModel(CompletionModel):
 
-    def __init__(self, parent):
-        CompletionModel.__init__(self, parent)
-        self.main_model = model = main.model()
+    def __init__(self, context, parent):
+        CompletionModel.__init__(self, context, parent)
+        self.main_model = model = context.model
         msg = model.message_updated
-        model.add_observer(msg, self.emit_update)
+        model.add_observer(msg, self.emit_model_updated)
 
     def gather_matches(self, case_sensitive):
         refs = filter_matches(self.match_text, self.matches(), case_sensitive,
                               sort_key=ref_sort_key)
         return (refs, (), set())
 
-    def emit_update(self):
+    def emit_model_updated(self):
         try:
-            self.emit(SIGNAL(UPDATE_SIGNAL))
-        except RuntimeError: # C++ object has been deleted
+            self.model_updated.emit()
+        except RuntimeError:  # C++ object has been deleted
             self.dispose()
 
     def matches(self):
         return []
 
     def dispose(self):
-        self.main_model.remove_observer(self.emit_update)
+        super(GitCompletionModel, self).dispose()
+        self.main_model.remove_observer(self.emit_model_updated)
 
 
 class GitRefCompletionModel(GitCompletionModel):
     """Completer for branches and tags"""
 
-    def __init__(self, parent):
-        GitCompletionModel.__init__(self, parent)
+    def __init__(self, context, parent):
+        GitCompletionModel.__init__(self, context, parent)
 
     def matches(self):
         model = self.main_model
         return model.local_branches + model.remote_branches + model.tags
 
 
-class GitPotentialBranchCompletionModel(GitCompletionModel):
-    """Completer for branches, tags, and potential branches"""
+def find_potential_branches(model):
+    remotes = model.remotes
+    remote_branches = model.remote_branches
 
-    def __init__(self, parent):
-        GitCompletionModel.__init__(self, parent)
+    ambiguous = set()
+    allnames = set(model.local_branches)
+    potential = []
+
+    for remote_branch in remote_branches:
+        branch = gitcmds.strip_remote(remotes, remote_branch)
+        if branch in allnames or branch == remote_branch:
+            ambiguous.add(branch)
+            continue
+        potential.append(branch)
+        allnames.add(branch)
+
+    potential_branches = [p for p in potential if p not in ambiguous]
+    return potential_branches
+
+
+class GitCreateBranchCompletionModel(GitCompletionModel):
+    """Completer for naming new branches"""
 
     def matches(self):
         model = self.main_model
-        remotes = model.remotes
-        remote_branches = model.remote_branches
+        potential_branches = find_potential_branches(model)
+        return (model.local_branches +
+                potential_branches +
+                model.tags)
 
-        ambiguous = set()
-        allnames = set(model.local_branches)
-        potential = []
 
-        for remote_branch in remote_branches:
-            branch = gitcmds.strip_remote(remotes, remote_branch)
-            if branch in allnames or branch == remote_branch:
-                ambiguous.add(branch)
-                continue
-            potential.append(branch)
-            allnames.add(branch)
-
-        potential_branches = [p for p in potential if p not in ambiguous]
+class GitCheckoutBranchCompletionModel(GitCompletionModel):
+    """Completer for git checkout <branch>"""
 
+    def matches(self):
+        model = self.main_model
+        potential_branches = find_potential_branches(model)
         return (model.local_branches +
                 potential_branches +
                 model.remote_branches +
@@ -517,10 +595,10 @@ class GitPotentialBranchCompletionModel(GitCompletionModel):
 
 
 class GitBranchCompletionModel(GitCompletionModel):
-    """Completer for remote branches"""
+    """Completer for local branches"""
 
-    def __init__(self, parent):
-        GitCompletionModel.__init__(self, parent)
+    def __init__(self, context, parent):
+        GitCompletionModel.__init__(self, context, parent)
 
     def matches(self):
         model = self.main_model
@@ -530,8 +608,8 @@ class GitBranchCompletionModel(GitCompletionModel):
 class GitRemoteBranchCompletionModel(GitCompletionModel):
     """Completer for remote branches"""
 
-    def __init__(self, parent):
-        GitCompletionModel.__init__(self, parent)
+    def __init__(self, context, parent):
+        GitCompletionModel.__init__(self, context, parent)
 
     def matches(self):
         model = self.main_model
@@ -541,8 +619,8 @@ class GitRemoteBranchCompletionModel(GitCompletionModel):
 class GitPathCompletionModel(GitCompletionModel):
     """Base class for path completion"""
 
-    def __init__(self, parent):
-        GitCompletionModel.__init__(self, parent)
+    def __init__(self, context, parent):
+        GitCompletionModel.__init__(self, context, parent)
 
     def candidate_paths(self):
         return []
@@ -557,8 +635,8 @@ class GitPathCompletionModel(GitCompletionModel):
 class GitStatusFilterCompletionModel(GitPathCompletionModel):
     """Completer for modified files and folders for status filtering"""
 
-    def __init__(self, parent):
-        GitPathCompletionModel.__init__(self, parent)
+    def __init__(self, context, parent):
+        GitPathCompletionModel.__init__(self, context, parent)
 
     def candidate_paths(self):
         model = self.main_model
@@ -569,15 +647,14 @@ class GitStatusFilterCompletionModel(GitPathCompletionModel):
 class GitTrackedCompletionModel(GitPathCompletionModel):
     """Completer for tracked files and folders"""
 
-    def __init__(self, parent):
-        GitPathCompletionModel.__init__(self, parent)
-        self.connect(self, SIGNAL(UPDATE_SIGNAL),
-                     self.gather_paths, Qt.QueuedConnection)
+    def __init__(self, context, parent):
+        GitPathCompletionModel.__init__(self, context, parent)
+        self.model_updated.connect(self.gather_paths, type=Qt.QueuedConnection)
         self._paths = []
-        self._updated = False
 
     def gather_paths(self):
-        self._paths = gitcmds.tracked_files()
+        context = self.context
+        self._paths = gitcmds.tracked_files(context)
 
     def gather_matches(self, case_sensitive):
         if not self._paths:
@@ -592,15 +669,14 @@ class GitTrackedCompletionModel(GitPathCompletionModel):
 class GitLogCompletionModel(GitRefCompletionModel):
     """Completer for arguments suitable for git-log like commands"""
 
-    def __init__(self, parent):
-        GitRefCompletionModel.__init__(self, parent)
-        self.connect(self, SIGNAL(UPDATE_SIGNAL),
-                     self.gather_paths, Qt.QueuedConnection)
+    def __init__(self, context, parent):
+        GitRefCompletionModel.__init__(self, context, parent)
+        self.model_updated.connect(self.gather_paths, type=Qt.QueuedConnection)
         self._paths = []
-        self._updated = False
 
     def gather_paths(self):
-        self._paths = gitcmds.tracked_files()
+        context = self.context
+        self._paths = gitcmds.tracked_files(context)
 
     def gather_matches(self, case_sensitive):
         if not self._paths:
@@ -620,44 +696,47 @@ class GitLogCompletionModel(GitRefCompletionModel):
         return (refs, paths, dirs)
 
 
-def bind_lineedit(model):
+def bind_lineedit(model, hint=''):
     """Create a line edit bound against a specific model"""
 
     class BoundLineEdit(CompletionLineEdit):
 
-        def __init__(self, hint='', parent=None):
-            CompletionLineEdit.__init__(self, model,
+        def __init__(self, context, hint=hint, parent=None):
+            CompletionLineEdit.__init__(self, context, model,
                                         hint=hint, parent=parent)
+            self.context = context
 
     return BoundLineEdit
 
 
 # Concrete classes
-GitLogLineEdit = bind_lineedit(GitLogCompletionModel)
-GitRefLineEdit = bind_lineedit(GitRefCompletionModel)
-GitPotentialBranchLineEdit = bind_lineedit(GitPotentialBranchCompletionModel)
-GitBranchLineEdit = bind_lineedit(GitBranchCompletionModel)
-GitRemoteBranchLineEdit = bind_lineedit(GitRemoteBranchCompletionModel)
-GitStatusFilterLineEdit = bind_lineedit(GitStatusFilterCompletionModel)
-GitTrackedLineEdit = bind_lineedit(GitTrackedCompletionModel)
+GitLogLineEdit = bind_lineedit(GitLogCompletionModel, hint='<ref>')
+GitRefLineEdit = bind_lineedit(GitRefCompletionModel, hint='<ref>')
+GitCheckoutBranchLineEdit = bind_lineedit(GitCheckoutBranchCompletionModel,
+                                          hint='<branch>')
+GitCreateBranchLineEdit = bind_lineedit(GitCreateBranchCompletionModel,
+                                        hint='<branch>')
+GitBranchLineEdit = bind_lineedit(GitBranchCompletionModel, hint='<branch>')
+GitRemoteBranchLineEdit = bind_lineedit(GitRemoteBranchCompletionModel,
+                                        hint='<remote-branch>')
+GitStatusFilterLineEdit = bind_lineedit(GitStatusFilterCompletionModel,
+                                        hint='<path>')
+GitTrackedLineEdit = bind_lineedit(GitTrackedCompletionModel, hint='<path>')
 
 
-class GitDialog(QtGui.QDialog):
+class GitDialog(QtWidgets.QDialog):
 
-    def __init__(self, lineedit, title, button_text, parent, icon=None):
-        QtGui.QDialog.__init__(self, parent)
+    def __init__(self, lineedit, context, title, text, parent, icon=None):
+        QtWidgets.QDialog.__init__(self, parent)
+        self.context = context
         self.setWindowTitle(title)
+        self.setWindowModality(Qt.WindowModal)
         self.setMinimumWidth(333)
 
-        self.label = QtGui.QLabel()
+        self.label = QtWidgets.QLabel()
         self.label.setText(title)
-
-        self.lineedit = lineedit()
-        self.setFocusProxy(self.lineedit)
-
-        if icon is None:
-            icon = icons.ok()
-        self.ok_button = qtutils.create_button(text=button_text, icon=icon)
+        self.lineedit = lineedit(context)
+        self.ok_button = qtutils.ok_button(text, icon=icon, enabled=False)
         self.close_button = qtutils.close_button()
 
         self.button_layout = qtutils.hbox(defs.no_margin, defs.button_spacing,
@@ -669,33 +748,30 @@ class GitDialog(QtGui.QDialog):
                                         self.button_layout)
         self.setLayout(self.main_layout)
 
+        self.lineedit.textChanged.connect(self.text_changed)
+        self.lineedit.enter.connect(self.accept)
         qtutils.connect_button(self.ok_button, self.accept)
         qtutils.connect_button(self.close_button, self.reject)
 
-        self.connect(self.lineedit, SIGNAL('textChanged(QString)'),
-                     self.text_changed)
-        self.connect(self.lineedit, SIGNAL('return()'), self.accept)
-
-        self.setWindowModality(Qt.WindowModal)
-        self.ok_button.setEnabled(False)
+        self.setFocusProxy(self.lineedit)
+        self.lineedit.setFocus(True)
 
     def text(self):
-        return ustr(self.lineedit.text())
+        return self.lineedit.text()
 
-    def text_changed(self, txt):
+    def text_changed(self, _txt):
         self.ok_button.setEnabled(bool(self.text()))
 
     def set_text(self, ref):
         self.lineedit.setText(ref)
 
     @classmethod
-    def get(cls, title, button_text, parent, default=None, icon=None):
-        dlg = cls(title, button_text, parent, icon=icon)
+    def get(cls, context, title, text, parent, default=None, icon=None):
+        dlg = cls(context, title, text, parent, icon=icon)
         if default:
             dlg.set_text(default)
 
         dlg.show()
-        dlg.raise_()
 
         def show_popup():
             x = dlg.lineedit.x()
@@ -705,38 +781,38 @@ class GitDialog(QtGui.QDialog):
             dlg.lineedit.popup().move(mapped.x(), mapped.y())
             dlg.lineedit.popup().show()
             dlg.lineedit.refresh()
+            dlg.lineedit.setFocus(True)
 
-        QtCore.QTimer().singleShot(0, show_popup)
+        QtCore.QTimer().singleShot(100, show_popup)
 
         if dlg.exec_() == cls.Accepted:
             return dlg.text()
-        else:
-            return None
+        return None
 
 
 class GitRefDialog(GitDialog):
 
-    def __init__(self, title, button_text, parent, icon=None):
-        GitDialog.__init__(self, GitRefLineEdit,
-                           title, button_text, parent, icon=icon)
+    def __init__(self, context, title, text, parent, icon=None):
+        GitDialog.__init__(
+            self, GitRefLineEdit, context, title, text, parent, icon=icon)
 
 
-class GitPotentialBranchDialog(GitDialog):
+class GitCheckoutBranchDialog(GitDialog):
 
-    def __init__(self, title, button_text, parent, icon=None):
-        GitDialog.__init__(self, GitPotentialBranchLineEdit,
-                           title, button_text, parent, icon=icon)
+    def __init__(self, context, title, text, parent, icon=None):
+        GitDialog.__init__(self, GitCheckoutBranchLineEdit,
+                           context, title, text, parent, icon=icon)
 
 
 class GitBranchDialog(GitDialog):
 
-    def __init__(self, title, button_text, parent, icon=None):
-        GitDialog.__init__(self, GitBranchLineEdit,
-                           title, button_text, parent, icon=icon)
+    def __init__(self, context, title, text, parent, icon=None):
+        GitDialog.__init__(
+            self, GitBranchLineEdit, context, title, text, parent, icon=icon)
 
 
 class GitRemoteBranchDialog(GitDialog):
 
-    def __init__(self, title, button_text, parent, icon=None):
+    def __init__(self, context, title, text, parent, icon=None):
         GitDialog.__init__(self, GitRemoteBranchLineEdit,
-                           title, button_text, parent, icon=icon)
+                           context, title, text, parent, icon=icon)
