@@ -1,32 +1,30 @@
 from __future__ import division, absolute_import, unicode_literals
+import os
 import re
 
+from qtpy import QtCore
 from qtpy import QtGui
 from qtpy import QtWidgets
 from qtpy.QtCore import Qt
 from qtpy.QtCore import Signal
 
 from ..i18n import N_
-from ..models import main
-from ..models import selection
-from ..qtutils import add_action
-from ..qtutils import create_action_button
-from ..qtutils import create_menu
-from ..qtutils import make_format
-from ..qtutils import RGB
+from ..interaction import Interaction
+from ..qtutils import get
 from .. import actions
 from .. import cmds
 from .. import core
 from .. import diffparse
-from .. import gitcfg
 from .. import gitcmds
 from .. import gravatar
 from .. import hotkeys
 from .. import icons
+from .. import utils
 from .. import qtutils
 from .text import TextDecorator
 from .text import VimHintedPlainTextEdit
 from . import defs
+from . import imageview
 
 
 COMMITS_SELECTED = 'COMMITS_SELECTED'
@@ -48,31 +46,32 @@ class DiffSyntaxHighlighter(QtGui.QSyntaxHighlighter):
                                       r'(?:@@@ (?:-[0-9,]+ ){2}\+[0-9,]+ @@@)')
     BAD_WHITESPACE_RGX = re.compile(r'\s+$')
 
-    def __init__(self, doc, whitespace=True, is_commit=False):
+    def __init__(self, context, doc, whitespace=True, is_commit=False):
         QtGui.QSyntaxHighlighter.__init__(self, doc)
         self.whitespace = whitespace
         self.enabled = True
         self.is_commit = is_commit
 
         QPalette = QtGui.QPalette
-        cfg = gitcfg.current()
+        cfg = context.cfg
         palette = QPalette()
         disabled = palette.color(QPalette.Disabled, QPalette.Text)
         header = qtutils.rgb_hex(disabled)
 
-        self.color_text = RGB(cfg.color('text', '030303'))
-        self.color_add = RGB(cfg.color('add', 'd2ffe4'))
-        self.color_remove = RGB(cfg.color('remove', 'fee0e4'))
-        self.color_header = RGB(cfg.color('header', header))
+        self.color_text = qtutils.RGB(cfg.color('text', '030303'))
+        self.color_add = qtutils.RGB(cfg.color('add', 'd2ffe4'))
+        self.color_remove = qtutils.RGB(cfg.color('remove', 'fee0e4'))
+        self.color_header = qtutils.RGB(cfg.color('header', header))
 
-        self.diff_header_fmt = make_format(fg=self.color_header)
-        self.bold_diff_header_fmt = make_format(fg=self.color_header, bold=True)
+        self.diff_header_fmt = qtutils.make_format(fg=self.color_header)
+        self.bold_diff_header_fmt = qtutils.make_format(
+            fg=self.color_header, bold=True)
 
-        self.diff_add_fmt = make_format(fg=self.color_text,
-                                        bg=self.color_add)
-        self.diff_remove_fmt = make_format(fg=self.color_text,
-                                           bg=self.color_remove)
-        self.bad_whitespace_fmt = make_format(bg=Qt.red)
+        self.diff_add_fmt = qtutils.make_format(
+            fg=self.color_text, bg=self.color_add)
+        self.diff_remove_fmt = qtutils.make_format(
+            fg=self.color_text, bg=self.color_remove)
+        self.bad_whitespace_fmt = qtutils.make_format(bg=Qt.red)
         self.setCurrentBlockState(self.INITIAL_STATE)
 
     def set_enabled(self, enabled):
@@ -134,18 +133,19 @@ class DiffSyntaxHighlighter(QtGui.QSyntaxHighlighter):
 
 class DiffTextEdit(VimHintedPlainTextEdit):
 
-    def __init__(self, parent,
+    def __init__(self, context, parent,
                  is_commit=False, whitespace=True, numbers=False):
-        VimHintedPlainTextEdit.__init__(self, '', parent=parent)
+        VimHintedPlainTextEdit.__init__(self, context, '', parent=parent)
         # Diff/patch syntax highlighter
-        self.highlighter = DiffSyntaxHighlighter(self.document(),
-                                                 is_commit=is_commit,
-                                                 whitespace=whitespace)
+        self.highlighter = DiffSyntaxHighlighter(
+            context, self.document(),
+            is_commit=is_commit, whitespace=whitespace)
         if numbers:
-            self.numbers = DiffLineNumbers(self)
+            self.numbers = DiffLineNumbers(context, self)
             self.numbers.hide()
         else:
             self.numbers = None
+        self.scrollvalue = None
 
         self.cursorPositionChanged.connect(self._cursor_changed)
 
@@ -160,27 +160,51 @@ class DiffTextEdit(VimHintedPlainTextEdit):
         if self.numbers:
             self.numbers.refresh_size()
 
+    def save_scrollbar(self):
+        """Save the scrollbar value, but only on the first call"""
+        if self.scrollvalue is None:
+            scrollbar = self.verticalScrollBar()
+            if scrollbar:
+                scrollvalue = get(scrollbar)
+            else:
+                scrollvalue = None
+            self.scrollvalue = scrollvalue
+
+    def restore_scrollbar(self):
+        """Restore the scrollbar and clear state"""
+        scrollbar = self.verticalScrollBar()
+        scrollvalue = self.scrollvalue
+        if scrollbar and scrollvalue is not None:
+            scrollbar.setValue(scrollvalue)
+        self.scrollvalue = None
+
     def set_loading_message(self):
         self.hint.set_value('+++ ' + N_('Loading...'))
         self.set_value('')
 
     def set_diff(self, diff):
+        """Set the diff text, but save the scrollbar"""
+        diff = diff.rstrip('\n')  # diffs include two empty newlines
+        self.save_scrollbar()
+
         self.hint.set_value('')
         if self.numbers:
             self.numbers.set_diff(diff)
         self.set_value(diff)
 
+        self.restore_scrollbar()
+
 
 class DiffLineNumbers(TextDecorator):
 
-    def __init__(self, parent):
+    def __init__(self, context, parent):
         TextDecorator.__init__(self, parent)
         self.highlight_line = -1
         self.lines = None
         self.parser = diffparse.DiffLines()
         self.formatter = diffparse.FormatDigits()
 
-        self.setFont(qtutils.diff_font())
+        self.setFont(qtutils.diff_font(context))
         self._char_width = self.fontMetrics().width('0')
 
         QPalette = QtGui.QPalette
@@ -237,19 +261,18 @@ class DiffLineNumbers(TextDecorator):
         editor = self.editor
         content_offset = editor.contentOffset()
         block = editor.firstVisibleBlock()
-        current_block_number = max(0, editor.textCursor().blockNumber())
         width = self.width()
         event_rect_bottom = event.rect().bottom()
 
         highlight = self._highlight
         highlight_text = self._highlight_text
-        window = self._window
         disabled = self._disabled
 
         fmt = self.formatter
         lines = self.lines
         num_lines = len(self.lines)
         painter.setPen(disabled)
+        text = ''
 
         while block.isValid():
             block_number = block.blockNumber()
@@ -262,13 +285,11 @@ class DiffLineNumbers(TextDecorator):
 
             rect = block_geom.translated(content_offset).toRect()
             if block_number == self.highlight_line:
-                painter.setPen(highlight_text)
                 painter.fillRect(rect.x(), rect.y(),
                                  width, rect.height(), highlight)
+                painter.setPen(highlight_text)
+            else:
                 painter.setPen(disabled)
-            elif block_number == current_block_number:
-                painter.fillRect(rect.x(), rect.y(),
-                                 width, rect.height(), window)
 
             line = lines[block_number]
             if len(line) == 2:
@@ -285,16 +306,289 @@ class DiffLineNumbers(TextDecorator):
             block = block.next()  # pylint: disable=next-method-called
 
 
-class DiffEditorWidget(QtWidgets.QWidget):
+class Viewer(QtWidgets.QWidget):
+    """Text and image diff viewers"""
 
-    def __init__(self, parent=None):
-        QtWidgets.QWidget.__init__(self, parent)
+    images_changed = Signal(object)
+    type_changed = Signal(object)
 
-        self.editor = DiffEditor(self, parent.titleBarWidget())
-        self.main_layout = qtutils.vbox(defs.no_margin, defs.spacing,
-                                        self.editor)
+    def __init__(self, context, parent=None):
+        super(Viewer, self).__init__(parent)
+
+        self.context = context
+        self.model = model = context.model
+        self.images = []
+        self.pixmaps = []
+        self.options = options = Options(self)
+        self.text = DiffEditor(context, options, self)
+        self.image = imageview.ImageView(parent=self)
+        self.image.setFocusPolicy(Qt.NoFocus)
+
+        stack = self.stack = QtWidgets.QStackedWidget(self)
+        stack.addWidget(self.text)
+        stack.addWidget(self.image)
+
+        self.main_layout = qtutils.vbox(
+            defs.no_margin, defs.no_spacing, self.stack)
         self.setLayout(self.main_layout)
-        self.setFocusProxy(self.editor)
+
+        # Observe images
+        images_msg = model.message_images_changed
+        model.add_observer(images_msg, self.images_changed.emit)
+        self.images_changed.connect(self.set_images, type=Qt.QueuedConnection)
+
+        # Observe the diff type
+        diff_type_msg = model.message_diff_type_changed
+        model.add_observer(diff_type_msg, self.type_changed.emit)
+        self.type_changed.connect(self.set_diff_type, type=Qt.QueuedConnection)
+
+        # Observe the image mode combo box
+        options.image_mode.currentIndexChanged.connect(lambda _: self.render())
+        options.zoom_mode.currentIndexChanged.connect(lambda _: self.render())
+
+        self.setFocusProxy(self.text)
+
+    def export_state(self, state):
+        state['show_diff_line_numbers'] = self.text.show_line_numbers()
+        state['image_diff_mode'] = self.options.image_mode.currentIndex()
+        state['image_zoom_mode'] = self.options.zoom_mode.currentIndex()
+        return state
+
+    def apply_state(self, state):
+        diff_numbers = bool(state.get('show_diff_line_numbers', False))
+        self.text.enable_line_numbers(diff_numbers)
+
+        image_mode = utils.asint(state.get('image_diff_mode', 0))
+        self.options.image_mode.set_index(image_mode)
+
+        zoom_mode = utils.asint(state.get('image_zoom_mode', 0))
+        self.options.zoom_mode.set_index(zoom_mode)
+        return True
+
+    def set_diff_type(self, diff_type):
+        """Manage the image and text diff views when selection changes"""
+        self.options.set_diff_type(diff_type)
+        if diff_type == 'image':
+            self.stack.setCurrentWidget(self.image)
+            self.render()
+        else:
+            self.stack.setCurrentWidget(self.text)
+
+    def update_options(self):
+        self.text.update_options()
+
+    def reset(self):
+        self.image.pixmap = QtGui.QPixmap()
+        self.cleanup()
+
+    def cleanup(self):
+        for (image, unlink) in self.images:
+            if unlink and core.exists(image):
+                os.unlink(image)
+        self.images = []
+
+    def set_images(self, images):
+        self.images = images
+        self.pixmaps = []
+        if not images:
+            self.reset()
+            return False
+
+        # In order to comp, we first have to load all the images
+        all_pixmaps = [QtGui.QPixmap(image[0]) for image in images]
+        pixmaps = [pixmap for pixmap in all_pixmaps if not pixmap.isNull()]
+        if not pixmaps:
+            self.reset()
+            return False
+
+        self.pixmaps = pixmaps
+        self.render()
+        self.cleanup()
+        return True
+
+    def render(self):
+        # Update images
+        if self.pixmaps:
+            mode = self.options.image_mode.currentIndex()
+            if mode == self.options.SIDE_BY_SIDE:
+                image = self.render_side_by_side()
+            elif mode == self.options.DIFF:
+                image = self.render_diff()
+            elif mode == self.options.XOR:
+                image = self.render_xor()
+            elif mode == self.options.PIXEL_XOR:
+                image = self.render_pixel_xor()
+            else:
+                image = self.render_side_by_side()
+        else:
+            image = QtGui.QPixmap()
+        self.image.pixmap = image
+
+        # Apply zoom
+        zoom_mode = self.options.zoom_mode.currentIndex()
+        zoom_factor = self.options.zoom_factors[zoom_mode][1]
+        if zoom_factor > 0.0:
+            self.image.resetTransform()
+            self.image.scale(zoom_factor, zoom_factor)
+            poly = self.image.mapToScene(self.image.viewport().rect())
+            self.image.last_scene_roi = poly.boundingRect()
+
+    def render_side_by_side(self):
+        # Side-by-side lineup comp
+        pixmaps = self.pixmaps
+        width = sum([pixmap.width() for pixmap in pixmaps])
+        height = max([pixmap.height() for pixmap in pixmaps])
+        image = create_image(width, height)
+
+        # Paint each pixmap
+        painter = create_painter(image)
+        x = 0
+        for pixmap in pixmaps:
+            painter.drawPixmap(x, 0, pixmap)
+            x += pixmap.width()
+        painter.end()
+
+        return image
+
+    def render_comp(self, comp_mode):
+        # Get the max size to use as the render canvas
+        pixmaps = self.pixmaps
+        if len(pixmaps) == 1:
+            return pixmaps[0]
+
+        width = max([pixmap.width() for pixmap in pixmaps])
+        height = max([pixmap.height() for pixmap in pixmaps])
+        image = create_image(width, height)
+
+        painter = create_painter(image)
+        for pixmap in (pixmaps[0], pixmaps[-1]):
+            x = (width - pixmap.width()) // 2
+            y = (height - pixmap.height()) // 2
+            painter.drawPixmap(x, y, pixmap)
+            painter.setCompositionMode(comp_mode)
+        painter.end()
+
+        return image
+
+    def render_diff(self):
+        comp_mode = QtGui.QPainter.CompositionMode_Difference
+        return self.render_comp(comp_mode)
+
+    def render_xor(self):
+        comp_mode = QtGui.QPainter.CompositionMode_Xor
+        return self.render_comp(comp_mode)
+
+    def render_pixel_xor(self):
+        comp_mode = QtGui.QPainter.RasterOp_SourceXorDestination
+        return self.render_comp(comp_mode)
+
+
+def create_image(width, height):
+    size = QtCore.QSize(width, height)
+    image = QtGui.QImage(size, QtGui.QImage.Format_ARGB32_Premultiplied)
+    image.fill(Qt.transparent)
+    return image
+
+
+def create_painter(image):
+    painter = QtGui.QPainter(image)
+    painter.fillRect(image.rect(), Qt.transparent)
+    return painter
+
+
+class Options(QtWidgets.QWidget):
+    """Provide the options widget used by the editor
+
+    Actions are registered on the parent widget.
+
+    """
+
+    # mode combobox indexes
+    SIDE_BY_SIDE = 0
+    DIFF = 1
+    XOR = 2
+    PIXEL_XOR = 3
+
+    def __init__(self, parent):
+        super(Options, self).__init__(parent)
+        self.widget = parent
+        self.ignore_space_at_eol = self.add_option(
+            N_('Ignore changes in whitespace at EOL'))
+
+        self.ignore_space_change = self.add_option(
+            N_('Ignore changes in amount of whitespace'))
+
+        self.ignore_all_space = self.add_option(
+            N_('Ignore all whitespace'))
+
+        self.function_context = self.add_option(
+            N_('Show whole surrounding functions of changes'))
+
+        self.show_line_numbers = self.add_option(
+            N_('Show line numbers'))
+
+        self.options = options = qtutils.create_action_button(
+            tooltip=N_('Diff Options'), icon=icons.configure())
+        qtutils.hide_button_menu_indicator(options)
+
+        self.image_mode = qtutils.combo([
+            N_('Side by side'),
+            N_('Diff'),
+            N_('XOR'),
+            N_('Pixel XOR'),
+        ])
+
+        self.zoom_factors = (
+            (N_('Zoom to Fit'), 0.0),
+            (N_('25%'), 0.25),
+            (N_('50%'), 0.5),
+            (N_('100%'), 1.0),
+            (N_('200%'), 2.0),
+            (N_('400%'), 4.0),
+            (N_('800%'), 8.0),
+        )
+        zoom_modes = [factor[0] for factor in self.zoom_factors]
+        self.zoom_mode = qtutils.combo(zoom_modes, parent=self)
+
+        self.menu = menu = qtutils.create_menu(N_('Diff Options'), options)
+        options.setMenu(menu)
+
+        menu.addAction(self.ignore_space_at_eol)
+        menu.addAction(self.ignore_space_change)
+        menu.addAction(self.ignore_all_space)
+        menu.addAction(self.show_line_numbers)
+        menu.addAction(self.function_context)
+
+        layout = qtutils.hbox(
+            defs.no_margin, defs.no_spacing,
+            self.image_mode, defs.button_spacing, self.zoom_mode, options)
+        self.setLayout(layout)
+
+        self.image_mode.setFocusPolicy(Qt.NoFocus)
+        self.zoom_mode.setFocusPolicy(Qt.NoFocus)
+        self.options.setFocusPolicy(Qt.NoFocus)
+        self.setFocusPolicy(Qt.NoFocus)
+
+    def set_diff_type(self, diff_type):
+        is_text = diff_type == 'text'
+        is_image = diff_type == 'image'
+        self.options.setVisible(is_text)
+        self.image_mode.setVisible(is_image)
+        self.zoom_mode.setVisible(is_image)
+
+    def add_option(self, title):
+        action = qtutils.add_action(self, title, self.update_options)
+        action.setCheckable(True)
+        return action
+
+    def update_options(self):
+        space_at_eol = get(self.ignore_space_at_eol)
+        space_change = get(self.ignore_space_change)
+        all_space = get(self.ignore_all_space)
+        function_context = get(self.function_context)
+        gitcmds.update_diff_overrides(
+            space_at_eol, space_change, all_space, function_context)
+        self.widget.update_options()
 
 
 class DiffEditor(DiffTextEdit):
@@ -303,52 +597,16 @@ class DiffEditor(DiffTextEdit):
     down = Signal()
     options_changed = Signal()
     updated = Signal()
-    diff_text_changed = Signal(object)
+    diff_text_updated = Signal(object)
 
-    def __init__(self, parent, titlebar):
-        DiffTextEdit.__init__(self, parent, numbers=True)
-        self.model = model = main.model()
+    def __init__(self, context, options, parent):
+        DiffTextEdit.__init__(self, context, parent, numbers=True)
+        self.context = context
+        self.model = model = context.model
+        self.selection_model = selection_model = context.selection
 
         # "Diff Options" tool menu
-        self.diff_ignore_space_at_eol_action = add_action(
-            self, N_('Ignore changes in whitespace at EOL'),
-            self._update_diff_opts)
-        self.diff_ignore_space_at_eol_action.setCheckable(True)
-
-        self.diff_ignore_space_change_action = add_action(
-            self, N_('Ignore changes in amount of whitespace'),
-            self._update_diff_opts)
-        self.diff_ignore_space_change_action.setCheckable(True)
-
-        self.diff_ignore_all_space_action = add_action(
-            self, N_('Ignore all whitespace'), self._update_diff_opts)
-        self.diff_ignore_all_space_action.setCheckable(True)
-
-        self.diff_function_context_action = add_action(
-            self, N_('Show whole surrounding functions of changes'),
-            self._update_diff_opts)
-        self.diff_function_context_action.setCheckable(True)
-
-        self.diff_show_line_numbers = add_action(
-            self, N_('Show lines numbers'),
-            self._update_diff_opts)
-        self.diff_show_line_numbers.setCheckable(True)
-
-        self.diffopts_button = create_action_button(
-            tooltip=N_('Diff Options'), icon=icons.configure())
-        self.diffopts_menu = create_menu(N_('Diff Options'),
-                                         self.diffopts_button)
-
-        self.diffopts_menu.addAction(self.diff_ignore_space_at_eol_action)
-        self.diffopts_menu.addAction(self.diff_ignore_space_change_action)
-        self.diffopts_menu.addAction(self.diff_ignore_all_space_action)
-        self.diffopts_menu.addAction(self.diff_show_line_numbers)
-        self.diffopts_menu.addAction(self.diff_function_context_action)
-        self.diffopts_button.setMenu(self.diffopts_menu)
-        qtutils.hide_button_menu_indicator(self.diffopts_button)
-
-        titlebar.add_corner_widget(self.diffopts_button)
-
+        self.options = options
         self.action_apply_selection = qtutils.add_action(
             self, 'Apply', self.apply_selection, hotkeys.STAGE_DIFF)
 
@@ -356,79 +614,74 @@ class DiffEditor(DiffTextEdit):
             self, 'Revert', self.revert_selection, hotkeys.REVERT)
         self.action_revert_selection.setIcon(icons.undo())
 
-        self.launch_editor = actions.launch_editor(self, *hotkeys.ACCEPT)
-        self.launch_difftool = actions.launch_difftool(self)
-        self.stage_or_unstage = actions.stage_or_unstage(self)
+        self.launch_editor = actions.launch_editor(context, self,
+                                                   *hotkeys.ACCEPT)
+        self.launch_difftool = actions.launch_difftool(context, self)
+        self.stage_or_unstage = actions.stage_or_unstage(context, self)
 
         # Emit up/down signals so that they can be routed by the main widget
         self.move_up = actions.move_up(self)
         self.move_down = actions.move_down(self)
 
-        diff_text_changed = model.message_diff_text_changed
-        model.add_observer(diff_text_changed, self.diff_text_changed.emit)
+        diff_text_updated = model.message_diff_text_updated
+        model.add_observer(diff_text_updated, self.diff_text_updated.emit)
+        self.diff_text_updated.connect(self.set_diff, type=Qt.QueuedConnection)
 
-        self.selection_model = selection_model = selection.selection_model()
-        selection_model.add_observer(selection_model.message_selection_changed,
-                                     self.updated.emit)
+        selection_model.add_observer(
+            selection_model.message_selection_changed, self.updated.emit)
         self.updated.connect(self.refresh, type=Qt.QueuedConnection)
 
-        self.diff_text_changed.connect(self.set_diff)
-
     def refresh(self):
         enabled = False
         s = self.selection_model.selection()
-        if s.modified and self.model.stageable():
-            if s.modified[0] in self.model.submodules:
+        model = self.model
+        if s.modified and model.stageable():
+            if s.modified[0] in model.submodules:
                 pass
-            elif s.modified[0] not in main.model().unstaged_deleted:
+            elif s.modified[0] not in model.unstaged_deleted:
                 enabled = True
         self.action_revert_selection.setEnabled(enabled)
 
     def enable_line_numbers(self, enabled):
         """Enable/disable the diff line number display"""
         self.numbers.setVisible(enabled)
-        self.diff_show_line_numbers.setChecked(enabled)
+        self.options.show_line_numbers.setChecked(enabled)
 
     def show_line_numbers(self):
         """Return True if we should show line numbers"""
-        return self.diff_show_line_numbers.isChecked()
+        return get(self.options.show_line_numbers)
 
-    def _update_diff_opts(self):
-        space_at_eol = self.diff_ignore_space_at_eol_action.isChecked()
-        space_change = self.diff_ignore_space_change_action.isChecked()
-        all_space = self.diff_ignore_all_space_action.isChecked()
-        function_context = self.diff_function_context_action.isChecked()
+    def update_options(self):
         self.numbers.setVisible(self.show_line_numbers())
-
-        gitcmds.update_diff_overrides(space_at_eol,
-                                      space_change,
-                                      all_space,
-                                      function_context)
         self.options_changed.emit()
 
     # Qt overrides
     def contextMenuEvent(self, event):
         """Create the context menu for the diff display."""
         menu = qtutils.create_menu(N_('Actions'), self)
-        s = selection.selection()
-        filename = selection.filename()
+        context = self.context
+        model = self.model
+        s = self.selection_model.selection()
+        filename = self.selection_model.filename()
 
-        if self.model.stageable() or self.model.unstageable():
-            if self.model.stageable():
+        if model.stageable() or model.unstageable():
+            if model.stageable():
                 self.stage_or_unstage.setText(N_('Stage'))
             else:
                 self.stage_or_unstage.setText(N_('Unstage'))
             menu.addAction(self.stage_or_unstage)
 
-        if s.modified and self.model.stageable():
-            if s.modified[0] in main.model().submodules:
-                action = menu.addAction(icons.add(), cmds.Stage.name(),
-                                        cmds.run(cmds.Stage, s.modified))
+        if s.modified and model.stageable():
+            item = s.modified[0]
+            if item in model.submodules:
+                path = core.abspath(item)
+                action = menu.addAction(
+                    icons.add(), cmds.Stage.name(),
+                    cmds.run(cmds.Stage, context, s.modified))
                 action.setShortcut(hotkeys.STAGE_SELECTION)
                 menu.addAction(icons.cola(), N_('Launch git-cola'),
-                               cmds.run(cmds.OpenRepo,
-                                        core.abspath(s.modified[0])))
-            elif s.modified[0] not in main.model().unstaged_deleted:
+                               cmds.run(cmds.OpenRepo, context, path))
+            elif item not in model.unstaged_deleted:
                 if self.has_selection():
                     apply_text = N_('Stage Selected Lines')
                     revert_text = N_('Revert Selected Lines...')
@@ -444,15 +697,17 @@ class DiffEditor(DiffTextEdit):
                 menu.addAction(self.action_apply_selection)
                 menu.addAction(self.action_revert_selection)
 
-        if s.staged and self.model.unstageable():
-            if s.staged[0] in main.model().submodules:
-                action = menu.addAction(icons.remove(), cmds.Unstage.name(),
-                                        cmds.do(cmds.Unstage, s.staged))
+        if s.staged and model.unstageable():
+            item = s.staged[0]
+            if item in model.submodules:
+                path = core.abspath(item)
+                action = menu.addAction(
+                    icons.remove(), cmds.Unstage.name(),
+                    cmds.run(cmds.Unstage, context, s.staged))
                 action.setShortcut(hotkeys.STAGE_SELECTION)
                 menu.addAction(icons.cola(), N_('Launch git-cola'),
-                               cmds.do(cmds.OpenRepo,
-                                       core.abspath(s.staged[0])))
-            elif s.staged[0] not in main.model().staged_deleted:
+                               cmds.run(cmds.OpenRepo, context, path))
+            elif item not in model.staged_deleted:
                 if self.has_selection():
                     apply_text = N_('Unstage Selected Lines')
                 else:
@@ -462,7 +717,7 @@ class DiffEditor(DiffTextEdit):
                 self.action_apply_selection.setIcon(icons.remove())
                 menu.addAction(self.action_apply_selection)
 
-        if self.model.stageable() or self.model.unstageable():
+        if model.stageable() or model.unstageable():
             # Do not show the "edit" action when the file does not exist.
             # Untracked files exist by definition.
             if filename and core.exists(filename):
@@ -496,64 +751,30 @@ class DiffEditor(DiffTextEdit):
                 cursor = self.cursorForPosition(event.pos())
                 self.setTextCursor(cursor)
 
-        return DiffTextEdit.mousePressEvent(self, event)
+        return super(DiffEditor, self).mousePressEvent(event)
 
     def setPlainText(self, text):
         """setPlainText(str) while retaining scrollbar positions"""
-        mode = self.model.mode
-        highlight = (mode != self.model.mode_none and
-                     mode != self.model.mode_untracked)
+        model = self.model
+        mode = model.mode
+        highlight = (mode != model.mode_none and
+                     mode != model.mode_untracked)
         self.highlighter.set_enabled(highlight)
 
         scrollbar = self.verticalScrollBar()
         if scrollbar:
-            scrollvalue = scrollbar.value()
+            scrollvalue = get(scrollbar)
         else:
             scrollvalue = None
 
         if text is None:
             return
 
-        offset, selection_text = self.offset_and_selection()
-        old_text = self.toPlainText()
-
         DiffTextEdit.setPlainText(self, text)
 
-        if selection_text and selection_text in text:
-            # If the old selection exists in the new text then re-select it.
-            idx = text.index(selection_text)
-            cursor = self.textCursor()
-            cursor.setPosition(idx)
-            cursor.setPosition(idx + len(selection_text),
-                               QtGui.QTextCursor.KeepAnchor)
-            self.setTextCursor(cursor)
-
-        elif text == old_text:
-            # Otherwise, if the text is identical and there is no selection
-            # then restore the cursor position.
-            cursor = self.textCursor()
-            cursor.setPosition(offset)
-            self.setTextCursor(cursor)
-        else:
-            # If none of the above applied then restore the cursor position.
-            position = max(0, min(offset, len(text) - 1))
-            cursor = self.textCursor()
-            cursor.setPosition(position)
-            cursor.movePosition(QtGui.QTextCursor.StartOfLine)
-            self.setTextCursor(cursor)
-
         if scrollbar and scrollvalue is not None:
             scrollbar.setValue(scrollvalue)
 
-    def has_selection(self):
-        return self.textCursor().hasSelection()
-
-    def offset_and_selection(self):
-        cursor = self.textCursor()
-        offset = cursor.selectionStart()
-        selection_text = cursor.selection().toPlainText()
-        return offset, selection_text
-
     def selected_lines(self):
         cursor = self.textCursor()
         selection_start = cursor.selectionStart()
@@ -564,7 +785,7 @@ class DiffEditor(DiffTextEdit):
         line_idx = 0
         line_start = 0
 
-        for line_idx, line in enumerate(self.value().splitlines()):
+        for line_idx, line in enumerate(get(self).splitlines()):
             line_end = line_start + len(line)
             if line_start <= selection_start <= line_end:
                 first_line_idx = line_idx
@@ -582,10 +803,11 @@ class DiffEditor(DiffTextEdit):
         return first_line_idx, last_line_idx
 
     def apply_selection(self):
-        s = selection.single_selection()
-        if self.model.stageable() and s.modified:
+        model = self.model
+        s = self.selection_model.single_selection()
+        if model.stageable() and s.modified:
             self.process_diff_selection()
-        elif self.model.unstageable():
+        elif model.unstageable():
             self.process_diff_selection(reverse=True)
 
     def revert_selection(self):
@@ -598,29 +820,33 @@ class DiffEditor(DiffTextEdit):
             title = N_('Revert Diff Hunk?')
             ok_text = N_('Revert Diff Hunk')
 
-        if not qtutils.confirm(title,
-                               N_('This operation drops uncommitted changes.\n'
-                                  'These changes cannot be recovered.'),
-                               N_('Revert the uncommitted changes?'),
-                               ok_text, default=True, icon=icons.undo()):
+        if not Interaction.confirm(
+                title,
+                N_('This operation drops uncommitted changes.\n'
+                   'These changes cannot be recovered.'),
+                N_('Revert the uncommitted changes?'),
+                ok_text, default=True, icon=icons.undo()):
             return
         self.process_diff_selection(reverse=True, apply_to_worktree=True)
 
     def process_diff_selection(self, reverse=False, apply_to_worktree=False):
         """Implement un/staging of the selected line(s) or hunk."""
-        if selection.selection_model().is_empty():
+        if self.selection_model.is_empty():
             return
+        context = self.context
         first_line_idx, last_line_idx = self.selected_lines()
-        cmds.do(cmds.ApplyDiffSelection, first_line_idx, last_line_idx,
+        cmds.do(cmds.ApplyDiffSelection, context,
+                first_line_idx, last_line_idx,
                 self.has_selection(), reverse, apply_to_worktree)
 
 
 class DiffWidget(QtWidgets.QWidget):
 
-    def __init__(self, notifier, parent, is_commit=False):
+    def __init__(self, context, notifier, parent, is_commit=False):
         QtWidgets.QWidget.__init__(self, parent)
 
-        self.runtask = qtutils.RunTask(parent=self)
+        self.context = context
+        self.oid = 'HEAD'
 
         author_font = QtGui.QFont(self.font())
         author_font.setPointSize(int(author_font.pointSize() * 1.1))
@@ -653,7 +879,8 @@ class DiffWidget(QtWidgets.QWidget):
         self.oid_label.setAlignment(Qt.AlignTop)
         self.oid_label.elide()
 
-        self.diff = DiffTextEdit(self, is_commit=is_commit, whitespace=False)
+        self.diff = DiffTextEdit(
+            context, self, is_commit=is_commit, whitespace=False)
 
         self.info_layout = qtutils.vbox(defs.no_margin, defs.no_spacing,
                                         self.author_label, self.summary_label,
@@ -670,11 +897,15 @@ class DiffWidget(QtWidgets.QWidget):
         notifier.add_observer(COMMITS_SELECTED, self.commits_selected)
         notifier.add_observer(FILES_SELECTED, self.files_selected)
 
+    def set_tabwidth(self, width):
+        self.diff.set_tabwidth(width)
+
     def set_diff_oid(self, oid, filename=None):
+        context = self.context
+        self.diff.save_scrollbar()
         self.diff.set_loading_message()
-        task = DiffInfoTask(oid, filename, self)
-        task.connect(self.diff.set_value)
-        self.runtask.start(task)
+        task = DiffInfoTask(context, oid, filename, self)
+        self.context.runtask.start(task, result=self.diff.set_diff)
 
     def commits_selected(self, commits):
         if len(commits) != 1:
@@ -687,9 +918,9 @@ class DiffWidget(QtWidgets.QWidget):
         author = commit.author or ''
 
         template_args = {
-                'author': author,
-                'email': email,
-                'summary': summary
+            'author': author,
+            'email': email,
+            'summary': summary
         }
 
         author_text = ("""%(author)s &lt;"""
@@ -762,10 +993,13 @@ class TextLabel(QtWidgets.QLabel):
 
 class DiffInfoTask(qtutils.Task):
 
-    def __init__(self, oid, filename, parent):
+    def __init__(self, context, oid, filename, parent):
         qtutils.Task.__init__(self, parent)
+        self.context = context
         self.oid = oid
         self.filename = filename
 
     def task(self):
-        return gitcmds.diff_info(self.oid, filename=self.filename)
+        context = self.context
+        oid = self.oid
+        return gitcmds.diff_info(context, oid, filename=self.filename)
