@@ -1,42 +1,38 @@
 from __future__ import division, absolute_import, unicode_literals
-
 import os
 
-from PyQt4 import QtCore
-from PyQt4 import QtGui
-from PyQt4.QtCore import SIGNAL
+from qtpy import QtCore
+from qtpy import QtWidgets
 
-from cola import cmds
-from cola import qtutils
-from cola import gitcfg
-from cola.i18n import N_
-from cola.models import prefs
-from cola.qtutils import diff_font
-from cola.widgets import defs
-from cola.widgets import standard
-from cola.compat import ustr
+from ..i18n import N_
+from ..models import prefs
+from .. import cmds
+from .. import qtutils
+from . import defs
+from . import standard
 
 
-def preferences(model=None, parent=None):
+def preferences(context, model=None, parent=None):
     if model is None:
-        model = prefs.PreferencesModel()
-    view = PreferencesView(model, parent=parent)
+        model = prefs.PreferencesModel(context)
+    view = PreferencesView(context, model, parent=parent)
     view.show()
     view.raise_()
     return view
 
 
-class FormWidget(QtGui.QWidget):
+class FormWidget(QtWidgets.QWidget):
 
-    def __init__(self, model, parent, source='user'):
-        QtGui.QWidget.__init__(self, parent)
+    def __init__(self, context, model, parent, source='user'):
+        QtWidgets.QWidget.__init__(self, parent)
+        self.context = context
+        self.cfg = context.cfg
         self.model = model
         self.config_to_widget = {}
         self.widget_to_config = {}
         self.source = source
-        self.config = gitcfg.current()
         self.defaults = {}
-        self.setLayout(QtGui.QFormLayout())
+        self.setLayout(QtWidgets.QFormLayout())
 
     def add_row(self, label, widget):
         self.layout().addRow(label, widget)
@@ -49,19 +45,17 @@ class FormWidget(QtGui.QWidget):
             self.connect_widget_to_config(widget, config)
 
     def connect_widget_to_config(self, widget, config):
-        if isinstance(widget, QtGui.QSpinBox):
-            widget.connect(widget, SIGNAL('valueChanged(int)'),
-                           self._int_config_changed(config))
+        if isinstance(widget, QtWidgets.QSpinBox):
+            widget.valueChanged.connect(self._int_config_changed(config))
 
-        elif isinstance(widget, QtGui.QCheckBox):
-            widget.connect(widget, SIGNAL('toggled(bool)'),
-                           self._bool_config_changed(config))
+        elif isinstance(widget, QtWidgets.QCheckBox):
+            widget.toggled.connect(self._bool_config_changed(config))
 
-        elif isinstance(widget, QtGui.QLineEdit):
-            widget.connect(widget, SIGNAL('editingFinished()'),
-                           self._text_config_changed(config))
-            widget.connect(widget, SIGNAL('returnPressed()'),
-                           self._text_config_changed(config))
+        elif isinstance(widget, QtWidgets.QLineEdit):
+            widget.editingFinished.connect(
+                self._text_config_changed(config, widget))
+            widget.returnPressed.connect(
+                self._text_config_changed(config, widget))
 
     def _int_config_changed(self, config):
         def runner(value):
@@ -73,17 +67,17 @@ class FormWidget(QtGui.QWidget):
             cmds.do(prefs.SetConfig, self.model, self.source, config, value)
         return runner
 
-    def _text_config_changed(self, config):
+    def _text_config_changed(self, config, widget):
         def runner():
-            value = ustr(self.sender().text())
+            value = widget.text()
             cmds.do(prefs.SetConfig, self.model, self.source, config, value)
         return runner
 
     def update_from_config(self):
         if self.source == 'user':
-            getter = self.config.get_user_or_system
+            getter = self.cfg.get_user_or_system
         else:
-            getter = self.config.get
+            getter = self.cfg.get
 
         for config, widget in self.widget_to_config.items():
             value = getter(config)
@@ -93,47 +87,47 @@ class FormWidget(QtGui.QWidget):
 
     def set_widget_value(self, widget, value):
         widget.blockSignals(True)
-        if isinstance(widget, QtGui.QSpinBox):
+        if isinstance(widget, QtWidgets.QSpinBox):
             widget.setValue(value)
-        elif isinstance(widget, QtGui.QLineEdit):
+        elif isinstance(widget, QtWidgets.QLineEdit):
             widget.setText(value)
-        elif isinstance(widget, QtGui.QCheckBox):
+        elif isinstance(widget, QtWidgets.QCheckBox):
             widget.setChecked(value)
         widget.blockSignals(False)
 
 
 class RepoFormWidget(FormWidget):
 
-    def __init__(self, model, parent, source):
-        FormWidget.__init__(self, model, parent, source=source)
+    def __init__(self, context, model, parent, source):
+        FormWidget.__init__(self, context, model, parent, source=source)
+        self.name = QtWidgets.QLineEdit()
+        self.email = QtWidgets.QLineEdit()
 
-        self.name = QtGui.QLineEdit()
-        self.email = QtGui.QLineEdit()
-        self.merge_verbosity = QtGui.QSpinBox()
-        self.merge_verbosity.setMinimum(0)
-        self.merge_verbosity.setMaximum(5)
-        self.merge_verbosity.setProperty('value', QtCore.QVariant(5))
-
-        self.diff_context = QtGui.QSpinBox()
-        self.diff_context.setMinimum(2)
-        self.diff_context.setMaximum(99)
-        self.diff_context.setProperty('value', QtCore.QVariant(5))
+        self.merge_verbosity = standard.SpinBox(value=5, maxi=5)
+        self.diff_context = standard.SpinBox(value=5, mini=2, maxi=9995)
 
         self.merge_summary = qtutils.checkbox(checked=True)
         self.merge_diffstat = qtutils.checkbox(checked=True)
         self.display_untracked = qtutils.checkbox(checked=True)
+        self.show_path = qtutils.checkbox(checked=True)
 
         tooltip = N_('Detect conflict markers in unmerged files')
         self.check_conflicts = qtutils.checkbox(checked=True, tooltip=tooltip)
 
+        tooltip = N_(
+            'Prevent "Stage" from staging all files when nothing is selected')
+        self.safe_mode = qtutils.checkbox(checked=False, tooltip=tooltip)
+
         self.add_row(N_('User Name'), self.name)
         self.add_row(N_('Email Address'), self.email)
         self.add_row(N_('Merge Verbosity'), self.merge_verbosity)
         self.add_row(N_('Number of Diff Context Lines'), self.diff_context)
         self.add_row(N_('Summarize Merge Commits'), self.merge_summary)
+        self.add_row(N_('Show Full Paths in the Window Title'), self.show_path)
         self.add_row(N_('Show Diffstat After Merge'), self.merge_diffstat)
         self.add_row(N_('Display Untracked Files'), self.display_untracked)
         self.add_row(N_('Detect Conflict Markers'), self.check_conflicts)
+        self.add_row(N_('Safe Mode'), self.safe_mode)
 
         self.set_config({
             prefs.CHECKCONFLICTS: (self.check_conflicts, True),
@@ -144,81 +138,84 @@ class RepoFormWidget(FormWidget):
             prefs.MERGE_DIFFSTAT: (self.merge_diffstat, True),
             prefs.MERGE_SUMMARY: (self.merge_summary, True),
             prefs.MERGE_VERBOSITY: (self.merge_verbosity, 5),
+            prefs.SAFE_MODE: (self.safe_mode, False),
+            prefs.SHOW_PATH: (self.show_path, True),
         })
 
 
 class SettingsFormWidget(FormWidget):
 
-    def __init__(self, model, parent):
-        FormWidget.__init__(self, model, parent)
+    def __init__(self, context, model, parent):
+        FormWidget.__init__(self, context, model, parent)
 
-        self.fixed_font = QtGui.QFontComboBox()
-        self.fixed_font.setFontFilters(QtGui.QFontComboBox.MonospacedFonts)
+        self.fixed_font = QtWidgets.QFontComboBox()
+        self.font_size = standard.SpinBox(value=12, mini=8, maxi=192)
 
-        self.font_size = QtGui.QSpinBox()
-        self.font_size.setMinimum(8)
-        self.font_size.setProperty('value', QtCore.QVariant(12))
-        self._font_str = None
+        self.maxrecent = standard.SpinBox(maxi=99)
+        self.tabwidth = standard.SpinBox(maxi=42)
+        self.textwidth = standard.SpinBox(maxi=150)
 
-        self.tabwidth = QtGui.QSpinBox()
-        self.tabwidth.setWrapping(True)
-        self.tabwidth.setMaximum(42)
-
-        self.textwidth = QtGui.QSpinBox()
-        self.textwidth.setWrapping(True)
-        self.textwidth.setMaximum(150)
+        self.editor = QtWidgets.QLineEdit()
+        self.historybrowser = QtWidgets.QLineEdit()
+        self.blameviewer = QtWidgets.QLineEdit()
+        self.difftool = QtWidgets.QLineEdit()
+        self.mergetool = QtWidgets.QLineEdit()
 
         self.linebreak = qtutils.checkbox()
-        self.editor = QtGui.QLineEdit()
-        self.historybrowser = QtGui.QLineEdit()
-        self.difftool = QtGui.QLineEdit()
-        self.mergetool = QtGui.QLineEdit()
         self.keep_merge_backups = qtutils.checkbox()
         self.sort_bookmarks = qtutils.checkbox()
         self.bold_headers = qtutils.checkbox()
         self.save_gui_settings = qtutils.checkbox()
+        self.check_spelling = qtutils.checkbox()
+        self.expandtab = qtutils.checkbox()
 
         self.add_row(N_('Fixed-Width Font'), self.fixed_font)
         self.add_row(N_('Font Size'), self.font_size)
         self.add_row(N_('Tab Width'), self.tabwidth)
+        self.add_row(N_('Insert spaces instead of tabs'), self.expandtab)
         self.add_row(N_('Text Width'), self.textwidth)
         self.add_row(N_('Auto-Wrap Lines'), self.linebreak)
         self.add_row(N_('Editor'), self.editor)
         self.add_row(N_('History Browser'), self.historybrowser)
+        self.add_row(N_('Blame Viewer'), self.blameviewer)
         self.add_row(N_('Diff Tool'), self.difftool)
         self.add_row(N_('Merge Tool'), self.mergetool)
-        self.add_row(N_('Keep *.orig Merge Backups'), self.keep_merge_backups)
+        self.add_row(N_('Recent repository count'), self.maxrecent)
         self.add_row(N_('Sort bookmarks alphabetically'), self.sort_bookmarks)
+        self.add_row(N_('Keep *.orig Merge Backups'), self.keep_merge_backups)
         self.add_row(N_('Bold with dark background font instead of italic '
                         'headers (restart required)'), self.bold_headers)
         self.add_row(N_('Save GUI Settings'), self.save_gui_settings)
+        self.add_row(N_('Check spelling'), self.check_spelling)
 
         self.set_config({
             prefs.SAVEWINDOWSETTINGS: (self.save_gui_settings, True),
             prefs.TABWIDTH: (self.tabwidth, 8),
+            prefs.EXPANDTAB: (self.expandtab, False),
             prefs.TEXTWIDTH: (self.textwidth, 72),
             prefs.LINEBREAK: (self.linebreak, True),
+            prefs.MAXRECENT: (self.maxrecent, 8),
             prefs.SORT_BOOKMARKS: (self.sort_bookmarks, True),
             prefs.BOLD_HEADERS: (self.bold_headers, False),
             prefs.DIFFTOOL: (self.difftool, 'xxdiff'),
             prefs.EDITOR: (self.editor, os.getenv('VISUAL', 'gvim')),
             prefs.HISTORY_BROWSER: (self.historybrowser,
                                     prefs.default_history_browser()),
+            prefs.BLAME_VIEWER: (self.blameviewer,
+                                 prefs.default_blame_viewer()),
             prefs.MERGE_KEEPBACKUP: (self.keep_merge_backups, True),
             prefs.MERGETOOL: (self.mergetool, 'xxdiff'),
+            prefs.SPELL_CHECK: (self.check_spelling, False),
         })
 
-        self.connect(self.fixed_font, SIGNAL('currentFontChanged(QFont)'),
-                     self.current_font_changed)
-
-        self.connect(self.font_size, SIGNAL('valueChanged(int)'),
-                     self.font_size_changed)
+        self.fixed_font.currentFontChanged.connect(self.current_font_changed)
+        self.font_size.valueChanged.connect(self.font_size_changed)
 
     def update_from_config(self):
         FormWidget.update_from_config(self)
-
+        context = self.context
         block = self.fixed_font.blockSignals(True)
-        font = diff_font()
+        font = qtutils.diff_font(context)
         self.fixed_font.setCurrentFont(font)
         self.fixed_font.blockSignals(block)
 
@@ -231,34 +228,35 @@ class SettingsFormWidget(FormWidget):
         font = self.fixed_font.currentFont()
         font.setPointSize(size)
         cmds.do(prefs.SetConfig, self.model,
-                'user', prefs.FONTDIFF, ustr(font.toString()))
+                'user', prefs.FONTDIFF, font.toString())
 
     def current_font_changed(self, font):
         cmds.do(prefs.SetConfig, self.model,
-                'user', prefs.FONTDIFF, ustr(font.toString()))
+                'user', prefs.FONTDIFF, font.toString())
 
 
 class PreferencesView(standard.Dialog):
 
-    def __init__(self, model, parent=None):
+    def __init__(self, context, model, parent=None):
         standard.Dialog.__init__(self, parent=parent)
+        self.context = context
         self.setWindowTitle(N_('Preferences'))
         if parent is not None:
             self.setWindowModality(QtCore.Qt.WindowModal)
 
         self.resize(600, 360)
 
-        self.tab_bar = QtGui.QTabBar()
+        self.tab_bar = QtWidgets.QTabBar()
         self.tab_bar.setDrawBase(False)
         self.tab_bar.addTab(N_('All Repositories'))
         self.tab_bar.addTab(N_('Current Repository'))
         self.tab_bar.addTab(N_('Settings'))
 
-        self.user_form = RepoFormWidget(model, self, source='user')
-        self.repo_form = RepoFormWidget(model, self, source='repo')
-        self.options_form = SettingsFormWidget(model, self)
+        self.user_form = RepoFormWidget(context, model, self, source='user')
+        self.repo_form = RepoFormWidget(context, model, self, source='repo')
+        self.options_form = SettingsFormWidget(context, model, self)
 
-        self.stack_widget = QtGui.QStackedWidget()
+        self.stack_widget = QtWidgets.QStackedWidget()
         self.stack_widget.addWidget(self.user_form)
         self.stack_widget.addWidget(self.repo_form)
         self.stack_widget.addWidget(self.options_form)
@@ -273,11 +271,8 @@ class PreferencesView(standard.Dialog):
                                         self.button_layout)
         self.setLayout(self.main_layout)
 
-        self.connect(self.tab_bar, SIGNAL('currentChanged(int)'),
-                     self.stack_widget.setCurrentIndex)
-
-        self.connect(self.stack_widget, SIGNAL('currentChanged(int)'),
-                     self.update_widget)
+        self.tab_bar.currentChanged.connect(self.stack_widget.setCurrentIndex)
+        self.stack_widget.currentChanged.connect(self.update_widget)
 
         qtutils.connect_button(self.close_button, self.accept)
         qtutils.add_close_action(self)
