@@ -1,5 +1,7 @@
 from __future__ import division, absolute_import, unicode_literals
 import itertools
+import os
+from functools import partial
 
 from qtpy.QtCore import Qt
 from qtpy.QtCore import Signal
@@ -7,19 +9,23 @@ from qtpy import QtCore
 from qtpy import QtWidgets
 
 from ..i18n import N_
-from ..models import main
 from ..models import prefs
 from ..models import selection
 from ..widgets import gitignore
+from ..widgets import standard
+from ..qtutils import get
+from .. import actions
 from .. import cmds
 from .. import core
 from .. import hotkeys
 from .. import icons
 from .. import qtutils
+from .. import settings
 from .. import utils
 from . import common
 from . import completion
 from . import defs
+from . import text
 
 
 class StatusWidget(QtWidgets.QWidget):
@@ -31,25 +37,25 @@ class StatusWidget(QtWidgets.QWidget):
 
     """
 
-    def __init__(self, titlebar, parent=None):
+    def __init__(self, context, titlebar, parent):
         QtWidgets.QWidget.__init__(self, parent)
+        self.context = context
 
         tooltip = N_('Toggle the paths filter')
         icon = icons.ellipsis()
         self.filter_button = qtutils.create_action_button(tooltip=tooltip,
                                                           icon=icon)
-        self.filter_widget = StatusFilterWidget()
+        self.filter_widget = StatusFilterWidget(context)
         self.filter_widget.hide()
-        self.tree = StatusTreeWidget()
+        self.tree = StatusTreeWidget(context, parent=self)
         self.setFocusProxy(self.tree)
 
         self.main_layout = qtutils.vbox(defs.no_margin, defs.no_spacing,
                                         self.filter_widget, self.tree)
         self.setLayout(self.main_layout)
 
-        self.toggle_action = qtutils.add_action(self, tooltip,
-                                                self.toggle_filter,
-                                                hotkeys.FILTER)
+        self.toggle_action = qtutils.add_action(
+            self, tooltip, self.toggle_filter, hotkeys.FILTER)
 
         titlebar.add_corner_widget(self.filter_button)
         qtutils.connect_button(self.filter_button, self.toggle_filter)
@@ -88,6 +94,7 @@ class StatusTreeWidget(QtWidgets.QTreeWidget):
     # Signals
     about_to_update = Signal()
     updated = Signal()
+    diff_text_changed = Signal()
 
     # Item categories
     idx_header = -1
@@ -100,8 +107,10 @@ class StatusTreeWidget(QtWidgets.QTreeWidget):
     # Read-only access to the mode state
     mode = property(lambda self: self.m.mode)
 
-    def __init__(self, parent=None):
+    def __init__(self, context, parent=None):
         QtWidgets.QTreeWidget.__init__(self, parent)
+        self.context = context
+        self.selection_model = context.selection
 
         self.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
         self.headerItem().setHidden(True)
@@ -128,36 +137,37 @@ class StatusTreeWidget(QtWidgets.QTreeWidget):
         self.old_selection = None
         self.old_contents = None
         self.old_current_item = None
+        self.was_visible = True
         self.expanded_items = set()
 
+        self.image_formats = qtutils.ImageFormats()
+
         self.process_selection_action = qtutils.add_action(
-            self, cmds.StageOrUnstage.name(),
-            cmds.run(cmds.StageOrUnstage), hotkeys.STAGE_SELECTION)
+            self, cmds.StageOrUnstage.name(), self.stage_selection,
+            hotkeys.STAGE_SELECTION)
 
         self.revert_unstaged_edits_action = qtutils.add_action(
             self, cmds.RevertUnstagedEdits.name(),
-            cmds.run(cmds.RevertUnstagedEdits), hotkeys.REVERT)
+            cmds.run(cmds.RevertUnstagedEdits, context), hotkeys.REVERT)
         self.revert_unstaged_edits_action.setIcon(icons.undo())
 
         self.launch_difftool_action = qtutils.add_action(
             self, cmds.LaunchDifftool.name(),
-            cmds.run(cmds.LaunchDifftool), hotkeys.DIFF)
+            cmds.run(cmds.LaunchDifftool, context), hotkeys.DIFF)
         self.launch_difftool_action.setIcon(icons.diff())
 
-        self.launch_editor_action = qtutils.add_action(
-            self, cmds.LaunchEditor.name(),
-            cmds.run(cmds.LaunchEditor), hotkeys.EDIT, *hotkeys.ACCEPT)
-        self.launch_editor_action.setIcon(icons.edit())
+        self.launch_editor_action = actions.launch_editor(
+            context, self, *hotkeys.ACCEPT)
 
         if not utils.is_win32():
             self.default_app_action = common.default_app_action(
-                    self, self.selected_group)
+                context, self, self.selected_group)
 
             self.parent_dir_action = common.parent_dir_action(
-                    self, self.selected_group)
+                context, self, self.selected_group)
 
             self.terminal_action = common.terminal_action(
-                    self, self.selected_group)
+                context, self, self.selected_group)
 
         self.up_action = qtutils.add_action(
             self, N_('Move Up'), self.move_up,
@@ -168,19 +178,43 @@ class StatusTreeWidget(QtWidgets.QTreeWidget):
             hotkeys.MOVE_DOWN, hotkeys.MOVE_DOWN_SECONDARY)
 
         self.copy_path_action = qtutils.add_action(
-            self, N_('Copy Path to Clipboard'), self.copy_path, hotkeys.COPY)
+            self, N_('Copy Path to Clipboard'),
+            partial(copy_path, context), hotkeys.COPY)
         self.copy_path_action.setIcon(icons.copy())
 
         self.copy_relpath_action = qtutils.add_action(
             self, N_('Copy Relative Path to Clipboard'),
-            self.copy_relpath, hotkeys.CUT)
+            partial(copy_relpath, context), hotkeys.CUT)
         self.copy_relpath_action.setIcon(icons.copy())
 
+        self.copy_leading_path_action = qtutils.add_action(
+            self, N_('Copy Leading Path to Clipboard'),
+            partial(copy_leading_path, context))
+        self.copy_leading_path_action.setIcon(icons.copy())
+
+        self.copy_basename_action = qtutils.add_action(
+            self, N_('Copy Basename to Clipboard'),
+            partial(copy_basename, context))
+        self.copy_basename_action.setIcon(icons.copy())
+
+        self.copy_customize_action = qtutils.add_action(
+            self, N_('Customize...'),
+            partial(customize_copy_actions, context, self))
+        self.copy_customize_action.setIcon(icons.configure())
+
         self.view_history_action = qtutils.add_action(
-            self, N_('View History...'), self.view_history, hotkeys.HISTORY)
+            self, N_('View History...'), partial(view_history, context),
+            hotkeys.HISTORY)
 
         self.view_blame_action = qtutils.add_action(
-            self, N_('Blame...'), self.view_blame, hotkeys.BLAME)
+            self, N_('Blame...'),
+            partial(view_blame, context), hotkeys.BLAME)
+
+        self.annex_add_action = qtutils.add_action(
+            self, N_('Add to Git Annex'), cmds.run(cmds.AnnexAdd, context))
+
+        self.lfs_track_action = qtutils.add_action(
+            self, N_('Add to Git LFS'), cmds.run(cmds.LFSTrack, context))
 
         # MoveToTrash and Delete use the same shortcut.
         # We will only bind one of them, depending on whether or not the
@@ -204,20 +238,30 @@ class StatusTreeWidget(QtWidgets.QTreeWidget):
         about_to_update = self._about_to_update
         self.about_to_update.connect(about_to_update, type=Qt.QueuedConnection)
         self.updated.connect(self.refresh, type=Qt.QueuedConnection)
+        self.diff_text_changed.connect(
+            self.make_current_item_visible, type=Qt.QueuedConnection)
 
-        self.m = main.model()
+        self.m = context.model
         self.m.add_observer(self.m.message_about_to_update,
                             self.about_to_update.emit)
         self.m.add_observer(self.m.message_updated, self.updated.emit)
+        self.m.add_observer(self.m.message_diff_text_changed,
+                            self.diff_text_changed.emit)
 
         self.itemSelectionChanged.connect(self.show_selection)
         self.itemDoubleClicked.connect(self.double_clicked)
         self.itemCollapsed.connect(lambda x: self.update_column_widths())
         self.itemExpanded.connect(lambda x: self.update_column_widths())
 
+    def make_current_item_visible(self):
+        item = self.currentItem()
+        if item:
+            self.scroll_to_item(item)
+
     def add_toplevel_item(self, txt, icon, hide=False):
+        context = self.context
         font = self.font()
-        if prefs.bold_headers():
+        if prefs.bold_headers(context):
             font.setBold(True)
         else:
             font.setItalic(True)
@@ -226,7 +270,7 @@ class StatusTreeWidget(QtWidgets.QTreeWidget):
         item.setFont(0, font)
         item.setText(0, txt)
         item.setIcon(0, icon)
-        if prefs.bold_headers():
+        if prefs.bold_headers(context):
             item.setBackground(0, self.palette().midlight())
         if hide:
             item.setHidden(True)
@@ -285,7 +329,7 @@ class StatusTreeWidget(QtWidgets.QTreeWidget):
             reselect = selection_info[3]
             try:
                 item = old[idx]
-            except:
+            except IndexError:
                 return
             if item in new:
                 reselect(item, current=True)
@@ -332,6 +376,25 @@ class StatusTreeWidget(QtWidgets.QTreeWidget):
             hscroll.setValue(self.old_hscroll)
             self.old_hscroll = None
 
+    def stage_selection(self):
+        """Stage or unstage files according to the selection"""
+        context = self.context
+        selected_indexes = self.selected_indexes()
+        if selected_indexes:
+            category, idx = selected_indexes[0]
+            # A header item e.g. 'Staged', 'Modified', etc.
+            if category == self.idx_header:
+                if idx == self.idx_staged:
+                    cmds.do(cmds.UnstageAll, context)
+                elif idx == self.idx_modified:
+                    cmds.do(cmds.StageModified, context)
+                elif idx == self.idx_untracked:
+                    cmds.do(cmds.StageUntracked, context)
+                else:
+                    pass  # Do nothing for unmerged items, by design
+                return
+        cmds.do(cmds.StageOrUnstage, context)
+
     def staged_item(self, itemidx):
         return self._subtree_item(self.idx_staged, itemidx)
 
@@ -374,11 +437,11 @@ class StatusTreeWidget(QtWidgets.QTreeWidget):
     def save_scrollbars(self):
         vscroll = self.verticalScrollBar()
         if vscroll:
-            self.old_vscroll = vscroll.value()
+            self.old_vscroll = get(vscroll)
 
         hscroll = self.horizontalScrollBar()
         if hscroll:
-            self.old_hscroll = hscroll.value()
+            self.old_hscroll = get(hscroll)
 
     def current_item(self):
         s = self.selected_indexes()
@@ -412,7 +475,7 @@ class StatusTreeWidget(QtWidgets.QTreeWidget):
 
     def update_actions(self, selected=None):
         if selected is None:
-            selected = selection.selection()
+            selected = self.selection_model.selection()
         can_revert_edits = bool(selected.staged or selected.modified)
         self.revert_unstaged_edits_action.setEnabled(can_revert_edits)
 
@@ -428,7 +491,9 @@ class StatusTreeWidget(QtWidgets.QTreeWidget):
 
     def set_unmerged(self, items):
         """Adds items to the 'Unmerged' subtree."""
-        self._set_subtree(items, self.idx_unmerged)
+        deleted_set = set([path for path in items if not core.exists(path)])
+        self._set_subtree(items, self.idx_unmerged,
+                          deleted_set=deleted_set)
 
     def set_untracked(self, items):
         """Adds items to the 'Untracked' subtree."""
@@ -487,8 +552,6 @@ class StatusTreeWidget(QtWidgets.QTreeWidget):
         """Set up the status menu for the repo status tree."""
         s = self.selection()
         menu = qtutils.create_menu('Status', self)
-        menu.addAction(self.launch_editor_action)
-
         selected_indexes = self.selected_indexes()
         if selected_indexes:
             category, idx = selected_indexes[0]
@@ -498,51 +561,94 @@ class StatusTreeWidget(QtWidgets.QTreeWidget):
 
         if s.staged:
             self._create_staged_context_menu(menu, s)
-
         elif s.unmerged:
             self._create_unmerged_context_menu(menu, s)
         else:
             self._create_unstaged_context_menu(menu, s)
 
         if not utils.is_win32():
-            menu.addSeparator()
-            menu.addAction(self.default_app_action)
-            menu.addAction(self.parent_dir_action)
+            if not menu.isEmpty():
+                menu.addSeparator()
+            if not self.selection_model.is_empty():
+                menu.addAction(self.default_app_action)
+                menu.addAction(self.parent_dir_action)
             menu.addAction(self.terminal_action)
 
+        self._add_copy_actions(menu)
+
         return menu
 
+    def _add_copy_actions(self, menu):
+        """Add the "Copy" sub-menu"""
+        enabled = self.selection_model.filename() is not None
+        self.copy_path_action.setEnabled(enabled)
+        self.copy_relpath_action.setEnabled(enabled)
+        self.copy_leading_path_action.setEnabled(enabled)
+        self.copy_basename_action.setEnabled(enabled)
+        copy_icon = icons.copy()
+
+        menu.addSeparator()
+        copy_menu = QtWidgets.QMenu(N_('Copy...'), menu)
+        menu.addMenu(copy_menu)
+
+        copy_menu.setIcon(copy_icon)
+        copy_menu.addAction(self.copy_path_action)
+        copy_menu.addAction(self.copy_relpath_action)
+        copy_menu.addAction(self.copy_leading_path_action)
+        copy_menu.addAction(self.copy_basename_action)
+
+        current_settings = settings.Settings()
+        current_settings.load()
+
+        copy_formats = current_settings.copy_formats
+        if copy_formats:
+            copy_menu.addSeparator()
+
+        context = self.context
+        for entry in copy_formats:
+            name = entry.get('name', '')
+            fmt = entry.get('format', '')
+            if name and fmt:
+                action = copy_menu.addAction(
+                    name, partial(copy_format, context, fmt))
+                action.setIcon(copy_icon)
+                action.setEnabled(enabled)
+
+        copy_menu.addSeparator()
+        copy_menu.addAction(self.copy_customize_action)
+
     def _create_header_context_menu(self, menu, idx):
+        context = self.context
         if idx == self.idx_staged:
             menu.addAction(icons.remove(), N_('Unstage All'),
-                           cmds.run(cmds.UnstageAll))
-            return menu
+                           cmds.run(cmds.UnstageAll, context))
         elif idx == self.idx_unmerged:
             action = menu.addAction(icons.add(), cmds.StageUnmerged.name(),
-                                    cmds.run(cmds.StageUnmerged))
+                                    cmds.run(cmds.StageUnmerged, context))
             action.setShortcut(hotkeys.STAGE_SELECTION)
-            return menu
         elif idx == self.idx_modified:
             action = menu.addAction(icons.add(), cmds.StageModified.name(),
-                                    cmds.run(cmds.StageModified))
+                                    cmds.run(cmds.StageModified, context))
             action.setShortcut(hotkeys.STAGE_SELECTION)
-            return menu
-
         elif idx == self.idx_untracked:
             action = menu.addAction(icons.add(), cmds.StageUntracked.name(),
-                                    cmds.run(cmds.StageUntracked))
+                                    cmds.run(cmds.StageUntracked, context))
             action.setShortcut(hotkeys.STAGE_SELECTION)
-            return menu
+        return menu
 
     def _create_staged_context_menu(self, menu, s):
         if s.staged[0] in self.m.submodules:
             return self._create_staged_submodule_context_menu(menu, s)
 
+        context = self.context
         if self.m.unstageable():
-            action = menu.addAction(icons.remove(), N_('Unstage Selected'),
-                                    cmds.run(cmds.Unstage, self.staged()))
+            action = menu.addAction(
+                icons.remove(), N_('Unstage Selected'),
+                cmds.run(cmds.Unstage, context, self.staged()))
             action.setShortcut(hotkeys.STAGE_SELECTION)
 
+        menu.addAction(self.launch_editor_action)
+
         # Do all of the selected items exist?
         all_exist = all(i not in self.m.staged_deleted and core.exists(i)
                         for i in self.staged())
@@ -553,53 +659,53 @@ class StatusTreeWidget(QtWidgets.QTreeWidget):
         if self.m.undoable():
             menu.addAction(self.revert_unstaged_edits_action)
 
-        menu.addSeparator()
-        menu.addAction(self.copy_path_action)
-        menu.addAction(self.copy_relpath_action)
         menu.addAction(self.view_history_action)
         menu.addAction(self.view_blame_action)
         return menu
 
     def _create_staged_submodule_context_menu(self, menu, s):
+        context = self.context
+        path = core.abspath(s.staged[0])
         menu.addAction(icons.cola(), N_('Launch git-cola'),
-                       cmds.run(cmds.OpenRepo,
-                                core.abspath(s.staged[0])))
-
-        action = menu.addAction(icons.remove(), N_('Unstage Selected'),
-                                cmds.run(cmds.Unstage, self.staged()))
+                       cmds.run(cmds.OpenRepo, context, path))
+        action = menu.addAction(
+            icons.remove(), N_('Unstage Selected'),
+            cmds.run(cmds.Unstage, context, self.staged()))
         action.setShortcut(hotkeys.STAGE_SELECTION)
-        menu.addSeparator()
 
-        menu.addAction(self.copy_path_action)
-        menu.addAction(self.copy_relpath_action)
         menu.addAction(self.view_history_action)
         return menu
 
-    def _create_unmerged_context_menu(self, menu, s):
+    def _create_unmerged_context_menu(self, menu, _s):
+        context = self.context
         menu.addAction(self.launch_difftool_action)
 
-        action = menu.addAction(icons.add(), N_('Stage Selected'),
-                                cmds.run(cmds.Stage, self.unstaged()))
+        action = menu.addAction(
+            icons.add(), N_('Stage Selected'),
+            cmds.run(cmds.Stage, context, self.unstaged()))
         action.setShortcut(hotkeys.STAGE_SELECTION)
 
-        menu.addSeparator()
-        menu.addAction(self.copy_path_action)
-        menu.addAction(self.copy_relpath_action)
+        menu.addAction(self.launch_editor_action)
         menu.addAction(self.view_history_action)
         menu.addAction(self.view_blame_action)
         return menu
 
     def _create_unstaged_context_menu(self, menu, s):
+        context = self.context
         modified_submodule = (s.modified and
                               s.modified[0] in self.m.submodules)
         if modified_submodule:
             return self._create_modified_submodule_context_menu(menu, s)
 
         if self.m.stageable():
-            action = menu.addAction(icons.add(), N_('Stage Selected'),
-                                    cmds.run(cmds.Stage, self.unstaged()))
+            action = menu.addAction(
+                icons.add(), N_('Stage Selected'),
+                cmds.run(cmds.Stage, context, self.unstaged()))
             action.setShortcut(hotkeys.STAGE_SELECTION)
 
+        if not self.selection_model.is_empty():
+            menu.addAction(self.launch_editor_action)
+
         # Do all of the selected items exist?
         all_exist = all(i not in self.m.unstaged_deleted and core.exists(i)
                         for i in self.staged())
@@ -613,62 +719,73 @@ class StatusTreeWidget(QtWidgets.QTreeWidget):
                 menu.addAction(self.revert_unstaged_edits_action)
 
         if all_exist and s.untracked:
+            # Git Annex / Git LFS
+            annex = self.m.annex
+            lfs = core.find_executable('git-lfs')
+            if annex or lfs:
+                menu.addSeparator()
+            if annex:
+                menu.addAction(self.annex_add_action)
+            if lfs:
+                menu.addAction(self.lfs_track_action)
+
             menu.addSeparator()
             if self.move_to_trash_action is not None:
                 menu.addAction(self.move_to_trash_action)
             menu.addAction(self.delete_untracked_files_action)
             menu.addSeparator()
-            menu.addAction(icons.edit(),
-                            N_('Add to .gitignore'),
-                            gitignore.gitignore_view)
-        menu.addSeparator()
-        menu.addAction(self.copy_path_action)
-        menu.addAction(self.copy_relpath_action)
-        if not selection.selection_model().is_empty():
+            menu.addAction(icons.edit(), N_('Add to .gitignore'),
+                           partial(gitignore.gitignore_view, self.context))
+
+        if not self.selection_model.is_empty():
             menu.addAction(self.view_history_action)
             menu.addAction(self.view_blame_action)
         return menu
 
     def _create_modified_submodule_context_menu(self, menu, s):
+        context = self.context
+        path = core.abspath(s.modified[0])
         menu.addAction(icons.cola(), N_('Launch git-cola'),
-                       cmds.run(cmds.OpenRepo, core.abspath(s.modified[0])))
+                       cmds.run(cmds.OpenRepo, context, path))
 
         if self.m.stageable():
             menu.addSeparator()
-            action = menu.addAction(icons.add(), N_('Stage Selected'),
-                                    cmds.run(cmds.Stage, self.unstaged()))
+            action = menu.addAction(
+                icons.add(), N_('Stage Selected'),
+                cmds.run(cmds.Stage, context, self.unstaged()))
             action.setShortcut(hotkeys.STAGE_SELECTION)
 
-        menu.addSeparator()
-        menu.addAction(self.copy_path_action)
-        menu.addAction(self.copy_relpath_action)
         menu.addAction(self.view_history_action)
         return menu
 
     def _delete_untracked_files(self):
-        cmds.do(cmds.Delete, self.untracked())
+        cmds.do(cmds.Delete, self.context, self.untracked())
 
     def _trash_untracked_files(self):
-        cmds.do(cmds.MoveToTrash, self.untracked())
+        cmds.do(cmds.MoveToTrash, self.context, self.untracked())
+
+    def selected_path(self):
+        s = self.single_selection()
+        return s.staged or s.unmerged or s.modified or s.untracked or None
 
     def single_selection(self):
         """Scan across staged, modified, etc. and return a single item."""
-        st = None
-        um = None
-        m = None
-        ut = None
+        staged = None
+        unmerged = None
+        modified = None
+        untracked = None
 
         s = self.selection()
         if s.staged:
-            st = s.staged[0]
-        elif s.modified:
-            m = s.modified[0]
+            staged = s.staged[0]
         elif s.unmerged:
-            um = s.unmerged[0]
+            unmerged = s.unmerged[0]
+        elif s.modified:
+            modified = s.modified[0]
         elif s.untracked:
-            ut = s.untracked[0]
+            untracked = s.untracked[0]
 
-        return selection.State(st, um, m, ut)
+        return selection.State(staged, unmerged, modified, untracked)
 
     def selected_indexes(self):
         """Returns a list of (category, row) representing the tree selection."""
@@ -698,17 +815,17 @@ class StatusTreeWidget(QtWidgets.QTreeWidget):
 
     def selected_group(self):
         """A list of selected files in various states of being"""
-        return selection.pick(self.selection())
+        return self.selection_model.pick(self.selection())
 
     def selected_idx(self):
         c = self.contents()
         s = self.single_selection()
         offset = 0
-        for content, selection in zip(c, s):
-            if len(content) == 0:
+        for content, sel in zip(c, s):
+            if not content:
                 continue
-            if selection is not None:
-                return offset + content.index(selection)
+            if sel is not None:
+                return offset + content.index(sel)
             offset += len(content)
         return None
 
@@ -721,25 +838,28 @@ class StatusTreeWidget(QtWidgets.QTreeWidget):
             (c.untracked, self.idx_untracked),
         ]
         for content, toplevel_idx in to_try:
-            if len(content) == 0:
+            if not content:
                 continue
             if idx < len(content):
                 parent = self.topLevelItem(toplevel_idx)
                 item = parent.child(idx)
-                self.select_item(item)
+                if item is not None:
+                    self.select_item(item)
                 return
             idx -= len(content)
 
-    def select_item(self, item):
+    def scroll_to_item(self, item):
         # First, scroll to the item, but keep the original hscroll
         hscroll = None
         hscrollbar = self.horizontalScrollBar()
         if hscrollbar:
-            hscroll = hscrollbar.value()
+            hscroll = get(hscrollbar)
         self.scrollToItem(item)
         if hscroll is not None:
             hscrollbar.setValue(hscroll)
 
+    def select_item(self, item):
+        self.scroll_to_item(item)
         self.setCurrentItem(item)
         item.setSelected(True)
 
@@ -782,27 +902,32 @@ class StatusTreeWidget(QtWidgets.QTreeWidget):
         item = self.topLevelItem(idx)
         return qtutils.tree_selection_items(item)
 
-    def double_clicked(self, item, idx):
+    def double_clicked(self, _item, _idx):
         """Called when an item is double-clicked in the repo status tree."""
-        cmds.do(cmds.StageOrUnstage)
+        cmds.do(cmds.StageOrUnstage, self.context)
 
     def show_selection(self):
         """Show the selected item."""
+        context = self.context
+        self.scroll_to_item(self.currentItem())
         # Sync the selection model
         selected = self.selection()
-        selection.selection_model().set_selection(selected)
+        selection_model = self.selection_model
+        selection_model.set_selection(selected)
         self.update_actions(selected=selected)
 
         selected_indexes = self.selected_indexes()
         if not selected_indexes:
             if self.m.amending():
-                cmds.do(cmds.SetDiffText, '')
+                cmds.do(cmds.SetDiffText, context, '')
             else:
-                cmds.do(cmds.ResetMode)
+                cmds.do(cmds.ResetMode, context)
             return
-        category, idx = selected_indexes[0]
+
         # A header item e.g. 'Staged', 'Modified', etc.
-        if category == self.idx_header:
+        category, idx = selected_indexes[0]
+        header = category == self.idx_header
+        if header:
             cls = {
                 self.idx_staged: cmds.DiffStagedSummary,
                 self.idx_modified: cmds.Diffstat,
@@ -810,24 +935,42 @@ class StatusTreeWidget(QtWidgets.QTreeWidget):
                 # self.idx_unmerged: cmds.UnmergedSummary,
                 self.idx_untracked: cmds.UntrackedSummary,
             }.get(idx, cmds.Diffstat)
-            cmds.do(cls)
-        # A staged file
-        elif category == self.idx_staged:
+            cmds.do(cls, context)
+            return
+
+        staged = category == self.idx_staged
+        modified = category == self.idx_modified
+        unmerged = category == self.idx_unmerged
+        untracked = category == self.idx_untracked
+
+        if staged:
             item = self.staged_items()[0]
-            cmds.do(cmds.DiffStaged, item.path, deleted=item.deleted)
-
-        # A modified file
-        elif category == self.idx_modified:
-            item = self.modified_items()[0]
-            cmds.do(cmds.Diff, item.path, deleted=item.deleted)
-
-        elif category == self.idx_unmerged:
+        elif unmerged:
             item = self.unmerged_items()[0]
-            cmds.do(cmds.Diff, item.path)
-
-        elif category == self.idx_untracked:
+        elif modified:
+            item = self.modified_items()[0]
+        elif untracked:
             item = self.unstaged_items()[0]
-            cmds.do(cmds.ShowUntracked, item.path)
+        else:
+            item = None  # this shouldn't happen
+        assert item is not None
+
+        path = item.path
+        deleted = item.deleted
+        image = self.image_formats.ok(path)
+
+        # Images are diffed differently
+        if image:
+            cmds.do(cmds.DiffImage, context, path, deleted,
+                    staged, modified, unmerged, untracked)
+        elif staged:
+            cmds.do(cmds.DiffStaged, context, path, deleted=deleted)
+        elif modified:
+            cmds.do(cmds.Diff, context, path, deleted=deleted)
+        elif unmerged:
+            cmds.do(cmds.Diff, context, path)
+        elif untracked:
+            cmds.do(cmds.ShowUntracked, context, path)
 
     def move_up(self):
         idx = self.selected_idx()
@@ -869,44 +1012,90 @@ class StatusTreeWidget(QtWidgets.QTreeWidget):
         else:
             self.select_by_index(0)
 
-    def copy_path(self, absolute=True):
-        """Copy a selected path to the clipboard"""
-        filename = selection.selection_model().filename()
-        qtutils.copy_path(filename, absolute=absolute)
-
-    def copy_relpath(self):
-        """Copy a selected relative path to the clipboard"""
-        self.copy_path(absolute=False)
-
     def _item_filter(self, item):
         return not item.deleted and core.exists(item.path)
 
     def mimeData(self, items):
         """Return a list of absolute-path URLs"""
+        context = self.context
         paths = qtutils.paths_from_items(items, item_filter=self._item_filter)
-        return qtutils.mimedata_from_paths(paths)
+        return qtutils.mimedata_from_paths(context, paths)
 
     def mimeTypes(self):
         return qtutils.path_mimetypes()
 
-    def view_blame(self):
-        """Signal that we should view blame for paths."""
-        cmds.do(cmds.BlamePaths, selection.union(selection.selection_model()))
 
-    def view_history(self):
-        """Signal that we should view history for paths."""
-        cmds.do(cmds.VisualizePaths,
-                selection.union(selection.selection_model()))
+def view_blame(context):
+    """Signal that we should view blame for paths."""
+    cmds.do(cmds.BlamePaths, context, context.selection.union())
+
+
+def view_history(context):
+    """Signal that we should view history for paths."""
+    cmds.do(cmds.VisualizePaths, context, context.selection.union())
+
+
+def copy_path(context, absolute=True):
+    """Copy a selected path to the clipboard"""
+    filename = context.selection.filename()
+    qtutils.copy_path(filename, absolute=absolute)
+
+
+def copy_relpath(context):
+    """Copy a selected relative path to the clipboard"""
+    copy_path(context, absolute=False)
+
+
+def copy_basename(context):
+    filename = os.path.basename(context.selection.filename())
+    basename, _ = os.path.splitext(filename)
+    qtutils.copy_path(basename, absolute=False)
+
+
+def copy_leading_path(context):
+    """Copy the selected leading path to the clipboard"""
+    filename = context.selection.filename()
+    dirname = os.path.dirname(filename)
+    qtutils.copy_path(dirname, absolute=False)
+
+
+def copy_format(context, fmt):
+    values = {}
+    values['path'] = path = context.selection.filename()
+    values['abspath'] = abspath = os.path.abspath(path)
+    values['absdirname'] = os.path.dirname(abspath)
+    values['dirname'] = os.path.dirname(path)
+    values['filename'] = os.path.basename(path)
+    values['basename'], values['ext'] = (
+        os.path.splitext(os.path.basename(path)))
+    qtutils.set_clipboard(fmt % values)
+
+
+def show_help(context):
+    help_text = N_(r"""
+        Format String Variables
+        -----------------------
+          %(path)s  =  relative file path
+       %(abspath)s  =  absolute file path
+       %(dirname)s  =  relative directory path
+    %(absdirname)s  =  absolute directory path
+      %(filename)s  =  file basename
+      %(basename)s  =  file basename without extension
+           %(ext)s  =  file extension
+""")
+    title = N_('Help - Custom Copy Actions')
+    return text.text_dialog(context, help_text, title)
 
 
 class StatusFilterWidget(QtWidgets.QWidget):
 
-    def __init__(self, parent=None):
+    def __init__(self, context, parent=None):
         QtWidgets.QWidget.__init__(self, parent)
-        self.main_model = main.model()
+        self.main_model = context.model
 
         hint = N_('Filter paths...')
-        self.text = completion.GitStatusFilterLineEdit(hint=hint, parent=self)
+        self.text = completion.GitStatusFilterLineEdit(
+            context, hint=hint, parent=self)
         self.text.setToolTip(hint)
         self.setFocusProxy(self.text)
         self._filter = None
@@ -914,16 +1103,148 @@ class StatusFilterWidget(QtWidgets.QWidget):
         self.main_layout = qtutils.hbox(defs.no_margin, defs.spacing, self.text)
         self.setLayout(self.main_layout)
 
-        text = self.text
-        text.changed.connect(self.apply_filter)
-        text.cleared.connect(self.apply_filter)
-        text.enter.connect(self.apply_filter)
-        text.editingFinished.connect(self.apply_filter)
+        widget = self.text
+        widget.changed.connect(self.apply_filter)
+        widget.cleared.connect(self.apply_filter)
+        widget.enter.connect(self.apply_filter)
+        widget.editingFinished.connect(self.apply_filter)
 
     def apply_filter(self):
-        text = self.text.value()
-        if text == self._filter:
+        value = get(self.text)
+        if value == self._filter:
             return
-        self._filter = text
-        paths = utils.shell_split(text)
+        self._filter = value
+        paths = utils.shell_split(value)
         self.main_model.update_path_filter(paths)
+
+
+def customize_copy_actions(context, parent):
+    """Customize copy actions"""
+    dialog = CustomizeCopyActions(context, parent)
+    dialog.show()
+    dialog.exec_()
+
+
+class CustomizeCopyActions(standard.Dialog):
+
+    def __init__(self, context, parent):
+        standard.Dialog.__init__(self, parent=parent)
+        self.setWindowTitle(N_('Custom Copy Actions'))
+
+        self.table = QtWidgets.QTableWidget(self)
+        self.table.setColumnCount(2)
+        self.table.setHorizontalHeaderLabels([
+            N_('Action Name'),
+            N_('Format String'),
+        ])
+        self.table.setSortingEnabled(False)
+        self.table.verticalHeader().hide()
+        self.table.horizontalHeader().setStretchLastSection(True)
+
+        self.add_button = qtutils.create_button(N_('Add'))
+        self.remove_button = qtutils.create_button(N_('Remove'))
+        self.remove_button.setEnabled(False)
+        self.show_help_button = qtutils.create_button(N_('Show Help'))
+        self.show_help_button.setShortcut(hotkeys.QUESTION)
+
+        self.close_button = qtutils.close_button()
+        self.save_button = qtutils.ok_button(N_('Save'))
+
+        self.buttons = qtutils.hbox(defs.no_margin, defs.button_spacing,
+                                    self.add_button,
+                                    self.remove_button,
+                                    self.show_help_button,
+                                    qtutils.STRETCH,
+                                    self.close_button,
+                                    self.save_button)
+
+        layout = qtutils.vbox(defs.margin, defs.spacing,
+                              self.table, self.buttons)
+        self.setLayout(layout)
+
+        qtutils.connect_button(self.add_button, self.add)
+        qtutils.connect_button(self.remove_button, self.remove)
+        qtutils.connect_button(
+            self.show_help_button, partial(show_help, context))
+        qtutils.connect_button(self.close_button, self.reject)
+        qtutils.connect_button(self.save_button, self.save)
+        qtutils.add_close_action(self)
+        self.table.itemSelectionChanged.connect(self.table_selection_changed)
+
+        self.init_size(parent=parent)
+
+        self.settings = settings.Settings()
+        QtCore.QTimer.singleShot(0, self.reload_settings)
+
+    def reload_settings(self):
+        # Called once after the GUI is initialized
+        self.settings.load()
+        table = self.table
+        for entry in self.settings.copy_formats:
+            name_string = entry.get('name', '')
+            format_string = entry.get('format', '')
+            if name_string and format_string:
+                name = QtWidgets.QTableWidgetItem(name_string)
+                fmt = QtWidgets.QTableWidgetItem(format_string)
+                rows = table.rowCount()
+                table.setRowCount(rows + 1)
+                table.setItem(rows, 0, name)
+                table.setItem(rows, 1, fmt)
+
+    def export_state(self):
+        state = super(CustomizeCopyActions, self).export_state()
+        standard.export_header_columns(self.table, state)
+        return state
+
+    def apply_state(self, state):
+        result = super(CustomizeCopyActions, self).apply_state(state)
+        standard.apply_header_columns(self.table, state)
+        return result
+
+    def add(self):
+        self.table.setFocus(True)
+        rows = self.table.rowCount()
+        self.table.setRowCount(rows + 1)
+
+        name = QtWidgets.QTableWidgetItem(N_('Name'))
+        fmt = QtWidgets.QTableWidgetItem(r'%(path)s')
+        self.table.setItem(rows, 0, name)
+        self.table.setItem(rows, 1, fmt)
+
+        self.table.setCurrentCell(rows, 0)
+        self.table.editItem(name)
+
+    def remove(self):
+        """Remove selected items"""
+        # Gather a unique set of rows and remove them in reverse order
+        rows = set()
+        items = self.table.selectedItems()
+        for item in items:
+            rows.add(self.table.row(item))
+
+        for row in reversed(sorted(rows)):
+            self.table.removeRow(row)
+
+    def save(self):
+        copy_formats = []
+        for row in range(self.table.rowCount()):
+            name = self.table.item(row, 0)
+            fmt = self.table.item(row, 1)
+            if name and fmt:
+                entry = {
+                    'name': name.text(),
+                    'format': fmt.text(),
+                }
+                copy_formats.append(entry)
+
+        while self.settings.copy_formats:
+            self.settings.copy_formats.pop()
+
+        self.settings.copy_formats.extend(copy_formats)
+        self.settings.save()
+
+        self.accept()
+
+    def table_selection_changed(self):
+        items = self.table.selectedItems()
+        self.remove_button.setEnabled(bool(items))
